"""."""

from kinematic_tracker.tracker.kkf import NdKkf


def test_nd_kkf_repr(kkf_wn: NdKkf) -> None:
    """."""
    ref = """NdKkf(NdKkfWhiteNoiseFd(ProcNoiseMeta(ProcNoiseKind.FINITE_DIFF
    factor 1.0
    mix_weight 0.75
    mult_by_last_derivative True
    is_dynamic_mixing True)))"""
    assert repr(kkf_wn) == ref



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_predict_all(tracker: NdKkfTracker) -> None:
    """."""
    tracker.predict_all(1234_000_000)
    assert tracker.last_ts_ns == 1234_000_000
    assert tracker.pre.last_dt == 1.234
    assert len(tracker.tracks) == 0



"""."""

import numpy as np

from kinematic_tracker.nd.precursors import NdKkfPrecursors


class NdKkfDiaNoise:
    """."""

    def __init__(self, factor: float) -> None:
        """Constructor.

        Args:
            factor: pre-factor to multiply the eye matrix.
        """
        self.factor = factor

    def __repr__(self) -> str:
        """."""
        return f'NdKkfDiaNoise({self.factor})'

    def fill_proc_cov(self, _pre: NdKkfPrecursors, _vec_x: np.ndarray, cov_xx: np.ndarray) -> None:
        """."""
        num_x = cov_xx.shape[0]
        cov_xx[:, :] = self.factor * np.eye(num_x)

    def save_values(self, vec_x: np.ndarray) -> None:
        """."""



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis


def test_compute_metric(
    det_rz: list[np.ndarray[int, float]], filters: list[cv2.KalmanFilter], driver: MetricMahalanobis
) -> None:
    metric = driver.compute_metric(det_rz, filters)
    ref = [
        [1.0, 0.9048374180359595],
        [0.9200444146293233, 1.0],
        [0.751477293075286, 0.9310627797040227],
    ]
    assert metric == pytest.approx(np.array(ref).T)
    assert np.shares_memory(metric, driver.metric_rt)


def test_wrong_inv(det_rz: list[np.ndarray[int, float]], driver: MetricMahalanobis) -> None:
    kf = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf.errorCovPre = np.ones((12, 12))
    kf.measurementMatrix = np.eye(6, 12)
    kf.measurementNoiseCov = np.zeros((6, 6))
    with pytest.raises(np.linalg.LinAlgError):
        driver.compute_metric(det_rz, [kf])



"""Generator of the block-diagonal, n-dimensional, kinematic, Kalman-filter matrices.

The fundamental- and process-noise matrices are generated for an n-dimensional state vector
composed of 1D kinematic state vectors.

The n-dimensional state vector is defined by a list of *kinematic orders* and their corresponding
*numbers of degrees of freedom* (DOF).
For example, a simple tracking_cuboid of cuboids with their centers processes with
constant-velocity (CV) point-motion model (PMM) and their dimensions processes with
constant-position (CP) PMM is defined by the orders [2, 1] and numbers of DOF [3, 3].
The distribution of variables in the composed state vector will be:

    state_vector = (pos_x, vel_x, pos_y, vel_y, pos_z, vel_z, dim_x, dim_y, dim_z)

To track the orientation (yaw) of cuboids we will add the 2-dimensional CP PMM to process
for x = cos(yaw) and y = sin(yaw). As a result the orders will be [2, 1, 1] and
numbers of DOF [3, 3, 2]. The distribution of variables in the composed state vector will be:

    state_vector = (pos_x, vel_x, pos_y, vel_y, pos_z, vel_z, dim_x, dim_y, dim_z, yaw_x, yaw_y)

Possible choices for kinematic orders range between 1 and 4. The correspondence between the
kinematic orders and PMMs is CP = 1, CV = 2, constant acceleration (CA) = 3 and
constant jerk (CJ) = 4. Number of DOF could vary from 1 to any big number. In practice,
the number of DOF is 2 or 3.

The number of variables in the state vector is  $N = sum_i Order_i * DOF_i$.
"""

from typing import Generator, Sequence

import numpy as np

from kinematic_tracker.matrices.factory import get_is_order_implemented

from .precursors import NdKkfPrecursors


def assert_compatible(orders: Sequence[int], num_dof: Sequence[int]) -> None:
    """Assert correctness of the kinematic orders and numbers of degrees of freedom (DoF).

    There should be at least one entry in the orders sequence.
    Every entry in orders should 1, 2, 3 or 4.
    Number of entries in orders should be the same as in num_dof.
    Every entry in num_dof should be positive.

    Args:
        orders: the kinematic orders for every part of the state vector.
        num_dof: numbers of degrees of freedom for every part of the state vector.

    Raises:
        Assert exceptions whenever some condition is not met.
    """
    assert len(orders) > 0
    assert all([get_is_order_implemented(o) for o in orders])
    assert len(orders) == len(num_dof)
    assert all([n > 0 for n in num_dof])


class NdKkfMatGenX:
    """Generator of the block-diagonal, n-dimensional, kinematic state matrices."""

    def __init__(self, orders: Sequence[int], num_dof: Sequence[int]) -> None:
        """Constructor

        Args:
            orders: the kinematic orders for every part of the state vector.
            num_dof: numbers of degrees of freedom for every part of the state vector.
        """
        assert_compatible(orders, num_dof)
        self.orders = np.empty(len(orders), dtype=int)
        self.orders_set = set()
        self.num_dof = np.array(num_dof, dtype=int)
        self.num_x = 0
        self.num_d = 0
        self.set_orders(orders)

    def set_orders(self, orders: Sequence[int]) -> None:
        """Setter for the kinematic orders.

        This is a helper method which allows to change the list of orders without
        creation of the new generator object.

        Args:
            orders: new kinematic orders.
        """
        assert_compatible(orders, self.num_dof)
        self.orders[:] = np.array(orders, dtype=int)
        self.orders_set.clear()
        self.orders_set |= set(orders)
        self.num_x = np.sum(self.orders * self.num_dof)
        self.num_d = int(self.num_dof.sum())

    def fill_q_mat(self, pre: NdKkfPrecursors, factors: np.ndarray, cov_xx: np.ndarray) -> None:
        """Fill the process-noise covariance matrix.

        Args:
            pre: a precursor object storing the precomputed 1D white-noise matrices.
            factors: a sequence of prefactors for each 1D part of the state vector.
            cov_xx: the resulting covariance matrix to be assigned (filled).

        Raises:
            AssertError: if precursor object or factor sequence are not compatible.
        """
        assert pre.orders_set == self.orders_set, (
            f' pre.orders_set {pre.orders_set} != gen.orders_set {self.orders_set}'
        )
        assert factors.shape == (self.num_d,)

        for (o, s, e), factor in zip(self.gen_loop(), factors):
            q_mat_dt = pre.kin_mat_gen[o].q_mat_dt
            cov_xx[s:e, s:e] = factor * q_mat_dt

    def fill_f_mat(self, pre: NdKkfPrecursors, f_mat_xx: np.ndarray) -> None:
        """Fill the transition (fundamental) matrix.

        Args:
            pre: a precursor object storing the precomputed 1D white-noise matrices.
            f_mat_xx: the resulting fundamental matrix to be assigned (filled).

        Raises:
            AssertError: if precursor object is not compatible with the generator.
        """
        assert pre.orders_set == self.orders_set
        for o, s, e in self.gen_loop():
            f_mat_dt = pre.kin_mat_gen[o].f_mat_dt
            f_mat_xx[s:e, s:e] = f_mat_dt

    def gen_loop(self) -> Generator[tuple[int, int, int], None, None]:
        """Generator of the loops over the 1D parts of the state vector.

        Returns:
            Generator yielding a complete sequence of orders, start and end indices.
        """
        s = 0
        for o, nd in zip(self.orders, self.num_dof):
            for _ix in range(nd):
                e = s + o
                yield o, s, e
                s = e



"""."""

from .kind import ProcNoiseKind


class ProcNoiseMeta:
    """."""

    def __init__(self) -> None:
        """."""
        self.noise_kind: ProcNoiseKind = ProcNoiseKind.FINITE_DIFF
        self.factor: float = 1.0
        self.mix_weight: float = 0.9
        self.mult_by_last_derivative: bool = True
        self.is_dynamic_mixing: bool = True

    def __repr__(self) -> str:
        """."""
        return (
            f'ProcNoiseMeta({self.noise_kind}\n'
            f'    factor {self.factor}\n'
            f'    mix_weight {self.mix_weight}\n'
            f'    mult_by_last_derivative {self.mult_by_last_derivative}\n'
            f'    is_dynamic_mixing {self.is_dynamic_mixing})'
        )


def get_proc_noise_meta(
    pn_kind: ProcNoiseKind,
    factor: float | None = None,
    mix_weight: float = 0.9,
    mult_by_last_der=True,
    is_dyn_mixing=True,
) -> ProcNoiseMeta:
    """A smarter than nothing factory for initialization of the process-noise-meta objects.

    Note: the expression sometimes below means "for some kinds of the process noise".

    Args:
        pn_kind: the process-noise kind. Only the user can decide this parameter.
        factor: the variance pre-factor in all cases.
                Sometimes I can use default value, sometimes not.
        mix_weight: the mixing weight used sometimes. I can always reasonably decide its value.
        mult_by_last_der: the flag used in the FINITE_DIFF process-noise kind, True by default.
        is_dyn_mixing: the flag used in the FINITE_DIFF process-noise kind, True by default.

    Returns:
        The object.

    Raises:
        ValueError: if the factor is None for the CONST or DIAGONAL process-noise kinds.
    """
    if pn_kind in (ProcNoiseKind.CONST, ProcNoiseKind.DIAGONAL) and factor is None:
        raise ValueError('I cannot reasonably decide the factor value')

    factor = 1.0 if factor is None else factor
    meta = ProcNoiseMeta()
    meta.noise_kind = pn_kind
    meta.factor = factor
    meta.mix_weight = mix_weight
    meta.mult_by_last_derivative = mult_by_last_der
    meta.is_dynamic_mixing = is_dyn_mixing
    return meta



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.match_driver_base import MatchDriverBase


def test_match_driver_base_init() -> None:
    """."""
    driver = MatchDriverBase(4, 5)
    assert driver.targets.shape == (4,)
    assert driver.reports.shape == (4,)


def test_match_driver_base_check_shape() -> None:
    """."""
    driver = MatchDriverBase(4, 5)
    driver.check_max_shape((2, 3))
    with pytest.raises(ValueError):
        driver.check_max_shape((5, 3))

    with pytest.raises(ValueError):
        driver.check_max_shape((2, 6))


def test_get_reports_targets_no_matches() -> None:
    """."""
    driver = MatchDriverBase(4, 5)
    reports, targets = driver.get_reports_targets()
    assert isinstance(reports, np.ndarray)
    assert isinstance(targets, np.ndarray)
    assert reports.ndim == 1
    assert targets.ndim == 1
    assert reports.size == 0
    assert targets.size == 0



"""."""

from pathlib import Path

import numpy as np

from kinematic_tracker.nd.derivative_sorted import DerivativeSorted
from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_cuboids_with_yaw() -> None:
    """."""

    det_path = Path(__file__).parent / 'annotation_task_1_757.csv'
    tss_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
    ids_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
    det_ly = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=range(2, 9))
    stamps = np.unique(tss_l)
    # Detections are prepared above.

    tracker = NdKkfTracker([3, 1, 1], [3, 3, 2])
    tracker.set_ind_pos_size((0, 1, 2, 3, 4, 5))
    tracker.set_measurement_std_dev(0, 0.1)  # define std dev for coordinates and...
    tracker.set_measurement_std_dev(1, 0.02)  # ... for dimensions of cuboids
    tracker.set_measurement_std_dev(2, 0.2)  # ... for yaw of cuboids
    # Tracker is initialized, tracking will be run below.

    der_sorted = DerivativeSorted(tracker.gen_xz.gen_x)
    buf_x = np.zeros((der_sorted.get_num_x(), 1))
    print()
    np.set_printoptions(linewidth=200, precision=4)
    for stamp_num, stamp_ns in enumerate(stamps):
        stamp_ns: int
        mask = tss_l == stamp_ns
        det_ry, ids_r = det_ly[mask], ids_l[mask]
        det_rz = convert_cartesian_yaw(det_ry)
        tracker.advance(stamp_ns, det_rz, ids_r)
        print(stamp_num)
        for track in tracker.tracks:
            der_sorted.convert_vec(track.kkf.kalman_filter.statePost, buf_x)
            print('  ', buf_x.T)  # tracking state in a derivative-sorted order:
            # px,py,py,dx,dy,dz,Psi_x,Psi_y,vx,vy,vz,ax,ay,ax


def convert_cartesian_yaw(det_ry: np.ndarray) -> np.ndarray[tuple[int, 8]]:
    """Linearize vectors with yaw angle.

    Essentially take the measurements
        y^T = (px, py, py, dx, dy, dz, Psi)
    and produce
        z^T = (px, py, py, dx, dy, dz, cos(Psi), sin(Psi))
    """
    assert det_ry.shape[1] == 7
    num_reports = len(det_ry)
    det_rz = np.zeros((num_reports, 8))
    det_rz[:, :6] = det_ry[:, :6]
    yaw_r = det_ry[:, 6]
    det_rz[:, 6] = np.cos(yaw_r)
    det_rz[:, 7] = np.sin(yaw_r)
    return det_rz



"""
This module defines the class WhiteNoiseMatCp, which computes and keeps a white-noise
covariance matrix for a constant position model. The covariance matrix is useful as
covariance of the process noise, typically after scaling by a variance prefactor.
"""

import numpy as np


class WhiteNoiseMatCp:
    """."""

    def __init__(self) -> None:
        """."""
        self.q_mat = np.zeros((1, 1))

    def compute(self, dt: float) -> None:
        """."""
        q_mat = self.q_mat
        q_mat[0, 0] = dt



"""."""

import numpy as np
import pytest

from cv2 import KalmanFilter


def test_kf_predict(kf_predicted: KalmanFilter) -> None:
    """."""
    x_ref = np.array([[7.0], [2.0], [15.0], [4.0], [5.0]])
    assert kf_predicted.statePre == pytest.approx(x_ref)
    assert kf_predicted.statePost == pytest.approx(x_ref)
    p_ref = np.array(
        [
            [957.0, 327.0, 0.0, 0.0, 0.0],
            [327.0, 118.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1911.0, 654.0, 0.0],
            [0.0, 0.0, 654.0, 236.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 57.0],
        ]
    )
    assert kf_predicted.errorCovPre == pytest.approx(p_ref)
    assert kf_predicted.errorCovPost == pytest.approx(p_ref)
    assert id(kf_predicted.errorCovPost) != id(kf_predicted.errorCovPre)



"""This is the transition matrix for the constant-acceleration point-motion model."""

import numpy as np


class FundMatCa:
    """."""

    def __init__(self) -> None:
        """."""
        self.f_mat = np.eye(3)

    def compute(self, dt: float) -> None:
        """."""
        f_mat = self.f_mat
        f_mat[0, 1] = dt
        f_mat[0, 2] = dt * dt / 2.0
        f_mat[1, 2] = dt



"""."""

import numpy as np

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors


class NdKkfWhiteNoiseConst:
    """."""

    def __init__(self, gen_x: NdKkfMatGenX, factor: float) -> None:
        """Constructor.

        Args:
            gen_x: driver of the matrix generation for kinematic states.
            factor: the pre-factor to multiply the eye.
        """
        self.gen_x = gen_x
        self.factor = factor
        self.factors = factor * np.ones(gen_x.num_d)

    def __repr__(self) -> str:
        """."""
        return f'NdKkfWhiteNoiseConst({self.factor})'

    def fill_proc_cov(self, pre: NdKkfPrecursors, _vec_x: np.ndarray, cov_xx: np.ndarray) -> None:
        """."""
        self.gen_x.fill_q_mat(pre, self.factors, cov_xx)

    def save_values(self, vec_x: np.ndarray) -> None:
        """."""



"""This is the transition matrix for the constant-jerk motion model for Cartesian point."""

import numpy as np


class FundMatCj:
    """."""

    def __init__(self) -> None:
        """."""
        self.f_mat = np.eye(4)

    def compute(self, dt: float) -> None:
        """."""
        sqr_dt2 = dt * dt / 2.0
        cub_dt6 = sqr_dt2 * dt / 3.0
        f_mat = self.f_mat
        f_mat[0, 1] = dt
        f_mat[0, 2] = sqr_dt2
        f_mat[0, 3] = cub_dt6
        f_mat[1, 2] = dt
        f_mat[1, 3] = sqr_dt2
        f_mat[2, 3] = dt



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.shape import NdKkfShape


def test_nd_kkf_init(shape: NdKkfShape) -> None:
    """."""
    assert isinstance(shape.orders_z, np.ndarray)
    assert isinstance(shape.orders_x, np.ndarray)
    assert isinstance(shape.num_dims, np.ndarray)
    assert shape.orders_z == pytest.approx([1, 2])
    assert shape.orders_x == pytest.approx([2, 3])
    assert shape.num_dims == pytest.approx([3, 2])


def test_nd_kkf_repr(shape: NdKkfShape) -> None:
    """."""
    assert repr(shape) == 'NdKkfShape( orders_x [2 3] num_dims [3 2] orders_z [1 2])'



"""Module for NdKkfWhiteNoiseFd.

This module defines the NdKkfWhiteNoiseFd class. The class serves as a driver
handling process-noise computations according to an original adaptive scheme.
It utilizes the white-noise template and dynamic factors to adjust the pre-factor
variance at each time step.

Classes:
    NdKkfWhiteNoiseFd: Handles process noise computations for kinematic states.
"""

import numpy as np

from kinematic_tracker.core.derivative_mixer import DerivativeMixer
from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors

from .meta import ProcNoiseMeta


class NdKkfWhiteNoiseFd:
    """
    Class for handling an adaptive process-noise calculation in a kinematic tracking system.

    Attributes:
        gen_x: Driver for the block-diagonal matrix machinery.
        pn_meta: Parameters for process noise computation.
        variance_factor: Static scaling factor for variance.
        der_mixer: Handles running averages of the derivatives absent in the state.
        dynamic_factors: Stores computed dynamic factors.
        tmp_last_vars: Temporary storage for last variables.
    """

    def __init__(self, gen_x: NdKkfMatGenX, pn_meta: ProcNoiseMeta, vec_x: np.ndarray) -> None:
        """
        Constructor for NdKkfWhiteNoiseFd.

        Args:
            gen_x: The driver used for block-diagonal distribution of variables.
            pn_meta: Parameters for computation of the process noise.
            vec_x: Initial state vector.
        """
        self.gen_x = gen_x
        self.pn_meta = pn_meta
        is_dyn = pn_meta.is_dynamic_mixing
        self.variance_factor = pn_meta.factor
        self.der_mixer = DerivativeMixer(pn_meta.mix_weight, gen_x.num_d, is_dyn)
        self.gather_last_vars(vec_x, self.der_mixer.values)
        self.dynamic_factors = np.zeros(gen_x.num_d)
        self.tmp_last_vars = np.zeros_like(self.der_mixer.values)

    def gather_last_vars(self, vec_x: np.ndarray, last_vars: np.ndarray) -> None:
        """
        Gathers the last variables from the state vector.

        Args:
            vec_x: State vector.
            last_vars: Storage for the last variables (highest-order derivatives).
        """
        self.assert_shape_x(vec_x)
        for i, (o, s, e) in enumerate(self.gen_x.gen_loop()):
            last_vars[i] = vec_x[e - 1]

    def assert_shape_x(self, vec_x: np.ndarray):
        """
        Asserts that the shape of the state vector is valid.

        Args:
            vec_x: State vector.

        Raises:
            AssertionError: If the shape of vec_x is not (gen_x.num_x,).
        """
        assert vec_x.shape == (self.gen_x.num_x,), (
            f'Need 1-D initial state vector, but got {vec_x.shape}'
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the object.

        Returns:
            str: String representation of the object.
        """
        return f'NdKkfWhiteNoiseFd({self.pn_meta})'

    def compute_dyn_factors(self, vec_x: np.ndarray, derivatives: np.ndarray) -> None:
        """
        Computes dynamic factors based on the state vector and derivatives.

        For each part of the state vector, it computes the dynamic factors by
        multiplying the last derivative with the next derivative. For example,
        if we deal with the constant velocity (CV) model, the dynamic factor will be

           $$
              \text{dynamic factor} = |a v|,
           $$
        where $a$ is the acceleration (running-average of the finite-differences
        of velocity), and $v$ is the velocity, i.e. the highest-order derivative
        available in the CV model.

        Args:
            vec_x: State vector.
            derivatives: Array of derivatives.
        """
        is_mult = self.pn_meta.mult_by_last_derivative
        for i, ((o, s, e), der) in enumerate(zip(self.gen_x.gen_loop(), derivatives)):
            last_var = vec_x[e - 1]
            self.dynamic_factors[i] = der * last_var if (o > 1 and is_mult) else der
        self.dynamic_factors *= self.variance_factor
        np.fabs(self.dynamic_factors, out=self.dynamic_factors)

    def fill_proc_cov(self, pre: NdKkfPrecursors, vec_x: np.ndarray, cov_xx: np.ndarray) -> None:
        """
        Fills the process noise covariance matrix.

        Args:
            pre: Precursor data with 1D white-noise matrices computed for a given time step.
            vec_x: State vector, posterior state vector in practice.
            cov_xx: The process-noise covariance matrix to be filled.
        """
        self.gather_last_vars(vec_x, self.tmp_last_vars)
        self.der_mixer.compute(pre.last_dt, self.tmp_last_vars)
        self.compute_dyn_factors(vec_x, self.der_mixer.derivatives)
        self.gen_x.fill_q_mat(pre, self.dynamic_factors, cov_xx)

    def save_values(self, vec_x: np.ndarray) -> None:
        """
        Saves the current state vector values for future computations.

        Args:
            vec_x: Input state vector, predicted or prior state in practice.
        """
        self.gather_last_vars(vec_x, self.der_mixer.values)



"""."""

import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


@pytest.fixture
def tracker() -> NdKkfTracker:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.set_measurement_std_dev(0, 0.02)
    tracker.set_measurement_std_dev(1, 0.001)
    tracker.set_association_metric('giou')
    tracker.association.threshold = 0.2
    tracker.num_det_min = 5
    tracker.num_misses_max = 5
    return tracker



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX


def test_nd_kkf_mat_gen_set_orders(gen_x: NdKkfMatGenX) -> None:
    """."""
    gen_x.set_orders([4, 3])
    assert gen_x.orders_set == {3, 4}
    assert gen_x.num_x == 18
    assert gen_x.num_d == 5


def test_nd_kkf_mat_set_wrong_orders(gen_x: NdKkfMatGenX) -> None:
    """."""
    with pytest.raises(AssertionError):
        gen_x.set_orders([4, 3, 1])
    with pytest.raises(AssertionError):
        gen_x.set_orders([4, 0])
    with pytest.raises(AssertionError):
        gen_x.set_orders([])



"""Module for the Hungarian bipartite matching.

This module provides a class `MatchDriverHungarian` that implements a necessary treatment of the
score matrix to be analysed with a linear-sum assignment (Hungarian) algorithm. It extends
the `MatchDriverBase` class. The linear-sum assignment from SciPy is applied.
"""

import numpy as np

from scipy.optimize import linear_sum_assignment

from .match_driver_base import MatchDriverBase


class MatchDriverHungarian(MatchDriverBase):
    """Class for Hungarian algorithm-based matching.

    This class uses the Hungarian algorithm to compute optimal matches
    between reports and targets based on a given score matrix.
    """

    def __init__(self, num_rows_max: int, num_cols_max: int) -> None:
        """
        Initialize the MatchDriverHungarian instance.

        This constructor initializes the base class and allocates a buffer for the score matrix.
        The buffer is necessary to because we have to treat the score matrix with a threshold
        before applying the linear-sum optimization (we do not want to change the input array
        of scores).

        Args:
            num_rows_max: Maximum number of rows in the score matrix.
            num_cols_max: Maximum number of columns in the score matrix.
        """
        super().__init__(num_rows_max, num_cols_max)
        self.score_rc = np.zeros((num_rows_max, num_cols_max))  # The buffer storage.

    def compute_matches(self, score: np.ndarray, threshold: float) -> None:
        """
        Compute the association map using the Hungarian algorithm.

        This method computes the optimal assignment of reports to targets
        based on the provided score matrix. Matches with scores below the
        threshold are ignored.

        Args:
            score (np.ndarray): The array of likelihoods (report, target) -> score.
            threshold (float): The threshold to declare the loose pairs.

        Returns:
            None: Updates the `reports` and `targets` attributes with the matches.
        """
        # Ensure the input score matrix does not exceed the pre-allocated size.
        self.check_max_shape(score.shape)

        # Reshape and copy the input score matrix into the pre-allocated matrix.
        rect_chunk = self.score_rc.reshape(-1)[: score.size].reshape(score.shape)
        rect_chunk[:] = score

        # Set scores below the threshold to zero.
        rect_chunk[rect_chunk < threshold] = 0.0

        # Apply the Hungarian algorithm to find the optimal assignment.
        rows, cols = linear_sum_assignment(rect_chunk, maximize=True)

        # Store the matches with score values above the threshold.
        self.num_matches = 0
        for row, col in zip(rows, cols):
            if score[row, col] >= threshold:
                self.reports[self.num_matches] = row
                self.targets[self.num_matches] = col
                self.num_matches += 1



"""."""

import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_set_association_threshold(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_association_threshold(0.4567)
    assert tracker.association.threshold == pytest.approx(0.4567)

    for invalid_value in (-1.0, 0.0, 1.0, 2.0):
        with pytest.raises(ValueError):
            tracker.set_association_threshold(invalid_value)



"""."""

import copy

import numpy as np
import pytest

from kinematic_tracker.association.g_iou_scores_aligned import GIoUAux


@pytest.fixture
def giou_aux() -> GIoUAux:
    """."""
    aux = GIoUAux()
    aux.set_vec_z(np.linspace(1.0, 6.0, num=6).reshape(6, 1))
    return aux


@pytest.fixture
def disjoint_xyz(giou_aux: GIoUAux) -> GIoUAux:
    """."""
    aux = copy.deepcopy(giou_aux)
    aux.corners += 10.0
    return aux


@pytest.fixture
def disjoint_x(giou_aux: GIoUAux) -> GIoUAux:
    """."""
    aux = copy.deepcopy(giou_aux)
    aux.corners[:, 0] += 10.0
    return aux


@pytest.fixture
def disjoint_xy(giou_aux: GIoUAux) -> GIoUAux:
    """."""
    aux = copy.deepcopy(giou_aux)
    aux.corners[:, 0:2] += 10.0
    return aux


@pytest.fixture
def joint_xy(giou_aux: GIoUAux) -> GIoUAux:
    """."""
    aux = copy.deepcopy(giou_aux)
    aux.corners[:, 0:2] += 1.0
    return aux



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_get_match(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_measurement_cov(np.eye(tracker.gen_xz.num_z))
    score_rt = tracker.metric_driver.compute_metric([np.ones(6)], [])
    tracker.match_driver.compute_matches(score_rt, tracker.association.threshold)
    assert tracker.match_driver.num_matches == 0


def test_nd_kkf_tracker_get_match1(tracker1: NdKkfTracker) -> None:
    """."""
    tracker1.predict_all(5234_000_000)
    score_rt = tracker1.metric_driver.compute_metric(
        [0.1 + np.linspace(1.0, 6.0, num=6)], tracker1.get_filters()
    )
    tracker1.match_driver.compute_matches(score_rt, tracker1.association.threshold)
    assert tracker1.match_driver.num_matches == 1
    assert score_rt == pytest.approx(0.84789335 * np.eye(1))



"""."""

import pytest

from kinematic_tracker.association.association import Association
from kinematic_tracker.association.association_metric import AssociationMetric
from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned
from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis
from kinematic_tracker.association.metric_size_modulated_mahalanobis import (
    MetricSizeModulatedMahalanobis,
)


def test_get_association_metric_driver(association: Association) -> None:
    """."""
    association.metric = AssociationMetric.MAHALANOBIS
    assert isinstance(association.get_metric_driver(), MetricMahalanobis)

    association.metric = AssociationMetric.GIOU
    assert isinstance(association.get_metric_driver(), MetricGIoUAligned)

    association.metric = AssociationMetric.SIZE_MODULATED_MAHALANOBIS
    assert isinstance(association.get_metric_driver(), MetricSizeModulatedMahalanobis)

    association.metric = AssociationMetric.UNKNOWN_METRIC
    with pytest.raises(ValueError):
        association.get_metric_driver()



"""Types of association metrics (matching scores)."""

from enum import Enum


class AssociationMetric(Enum):
    """."""

    GIOU = 'giou'
    MAHALANOBIS = 'mahalanobis'
    SIZE_MODULATED_MAHALANOBIS = 'size-modulated-mahalanobis'
    UNKNOWN_METRIC = 'unknown-metric'



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_set_measurement_cov_normal(tracker: NdKkfTracker) -> None:
    """."""
    assert tracker.cov_zz is None
    tracker.set_measurement_cov(0.678 * np.eye(6))
    assert tracker.cov_zz == pytest.approx(0.678 * np.eye(6))


def test_set_measurement_cov_wrong_shape(tracker: NdKkfTracker) -> None:
    """."""
    with pytest.raises(ValueError):
        tracker.set_measurement_cov(np.ones(6))

    with pytest.raises(ValueError):
        tracker.set_measurement_cov(np.eye(5))

    with pytest.raises(ValueError):
        tracker.set_measurement_cov(np.zeros((6, 6, 6)))



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.proc_noise.meta import ProcNoiseKind, ProcNoiseMeta, get_proc_noise_meta


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    """."""
    return NdKkfMatGenX([2, 1], [3, 2])


@pytest.fixture
def pn_meta_dia() -> ProcNoiseMeta:
    """."""
    return get_proc_noise_meta(ProcNoiseKind.DIAGONAL, 1.12, 0.75, True, True)



"""."""

import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_advance(tracker1: NdKkfTracker) -> None:
    """."""
    assert tracker1.last_ts_ns == 1234_000_000
    assert tracker1.pre.last_dt == pytest.approx(1.234)
    assert len(tracker1.tracks) == 1
    assert tracker1.tracks[0].ann_id == 123
    assert tracker1.tracks[0].upd_id == 123
    assert tracker1.tracks[0].num_miss == 0
    assert tracker1.tracks[0].num_det == 0
    vec_x_ref = [1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 5, 6]
    assert tracker1.tracks[0].kkf.kalman_filter.statePost[:, 0] == pytest.approx(vec_x_ref)



"""
This module defines the class WhiteNoiseMatCa, which computes and keeps a white-noise
covariance matrix for a constant-acceleration point-motion. The covariance matrix is useful as
covariance of the process noise, typically after scaling by a variance prefactor.
"""

import numpy as np


class WhiteNoiseMatCa:
    """."""

    def __init__(self) -> None:
        """."""
        self.q_mat = np.zeros((3, 3))

    def compute(self, dt: float) -> None:
        """."""
        mat = self.q_mat
        sqr = dt * dt
        cub = sqr * dt
        qua = cub * dt
        mat[0, :] = qua / 20, cub / 8, sqr / 6
        mat[1, :] = mat[0, 1], sqr / 3, dt / 2
        mat[2, :] = mat[0, 2], mat[1, 2], 1
        mat *= dt



"""."""

from kinematic_tracker.proc_noise.nd_kkf_dia_noise import NdKkfDiaNoise


def test_nd_kkf_dia_noise_repr(dia_n: NdKkfDiaNoise) -> None:
    """."""
    assert repr(dia_n) == 'NdKkfDiaNoise(5.678)'



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.proc_noise.nd_kkf_dia_noise import NdKkfDiaNoise


def test_fill_proc_cov_diagonal(dia_n: NdKkfDiaNoise, pre: NdKkfPrecursors) -> None:
    """."""
    cov_xx = 1.3 * np.ones((8, 8))
    pre.compute(9.0)
    dia_n.fill_proc_cov(pre, np.empty(8), cov_xx)
    assert cov_xx == pytest.approx(5.678 * np.eye(8))



"""
This module defines the class WhiteNoiseMatCj, which computes and keeps a white-noise
covariance matrix for a constant-jerk point-motion. The covariance matrix is useful as
covariance of the process noise, typically after scaling by a variance prefactor.
"""

import numpy as np


class WhiteNoiseMatCj:
    """."""

    def __init__(self) -> None:
        """."""
        self.q_mat = np.zeros((4, 4))

    def compute(self, dt: float) -> None:
        """."""
        q_mat = self.q_mat
        dt2 = dt * dt
        dt3 = dt2 * dt
        dt4 = dt3 * dt
        dt5 = dt4 * dt
        dt6 = dt5 * dt
        q_mat[0, 0] = dt6 / 252
        q_mat[0, 1] = dt5 / 72
        q_mat[0, 2] = dt4 / 30
        q_mat[0, 3] = dt3 / 24
        q_mat[1, 0] = q_mat[0, 1]
        q_mat[1, 1] = dt4 / 20
        q_mat[1, 2] = dt3 / 8
        q_mat[1, 3] = dt2 / 6
        q_mat[2, 0] = q_mat[0, 2]
        q_mat[2, 1] = q_mat[1, 2]
        q_mat[2, 2] = dt2 / 3
        q_mat[2, 3] = dt / 2
        q_mat[3, 0] = q_mat[0, 3]
        q_mat[3, 1] = q_mat[1, 3]
        q_mat[3, 2] = q_mat[2, 3]
        q_mat[3, 3] = 1.0
        q_mat *= dt



"""."""

from kinematic_tracker.proc_noise.meta import ProcNoiseMeta


def test_proc_noise_meta_repr(pn_meta_fd: ProcNoiseMeta) -> None:
    """."""
    ref = """ProcNoiseMeta(ProcNoiseKind.FINITE_DIFF
    factor 1.0
    mix_weight 0.75
    mult_by_last_derivative True
    is_dynamic_mixing True)"""
    assert repr(pn_meta_fd) == ref



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_promote_clutter1(tracker1: NdKkfTracker) -> None:
    """."""
    tracker1.tracks[0].num_det = 5
    tracker1.promote_clutter_eventually()
    assert tracker1.tracks[0].id == 0



"""."""

import pytest

from kinematic_tracker.proc_noise.kind import ProcNoiseKind
from kinematic_tracker.proc_noise.meta import get_proc_noise_meta


def test_corner_cases() -> None:
    """."""
    with pytest.raises(ValueError):
        get_proc_noise_meta(ProcNoiseKind.DIAGONAL)



"""
This module defines the `NdKkfTrack` class. The class represents a tracked target
used in a tracker based on kinematic Kalman filter.
It provides methods to manage the number of consecutive detections, misses.
"""

from association_quality_clavia import UPD_ID_LOOSE

from .kkf import NdKkf


class NdKkfTrack:
    """
    Represents a track in a kinematic tracking system using an NdKkf (N-dimensional Kalman filter).

    Attributes:
        kkf: The kinematic Kalman filter instance associated with this track.
        id: The locally unique identifier for the track. Defaults to -1 (unconfirmed object).
        creation_id: The ID associated with the creation of the track. Defaults to -1.
                     Used as a globally unique identifier for the targets.
                     The creation ID will be not repeated during the lifetime of the tracker.
        ann_id: The annotation ID associated with the track.
        upd_id: The ID of the last update. Defaults to the detection ID.
        score: The score of the track, representing its association metric value. Defaults to 0.0.
        num_det: The number of consecutive associations being done to this track. Defaults to 0.
        num_miss: The number of consecutive misses for the track. Defaults to 0.
    """

    def __init__(self, kkf: NdKkf, ann_id: int) -> None:
        """
        Initializes an instance of the `NdKkfTrack` class.

        Args:
            kkf: The kinematic Kalman filter instance to use in this track.
            ann_id: The annotation ID to mark the track with. The annotation ID
                    is not changing during the lifetime of the track.
        """
        self.kkf = kkf
        self.id = -1
        self.creation_id: int = -1
        self.ann_id = ann_id
        self.upd_id = ann_id
        self.score: float = 0.0
        self.num_det = 0
        self.num_miss = 0

    def bump_num_detections(self, score: float, upd_id: int) -> None:
        """
        Updates the consecutive detection/misses counters with a new detection.

        Args:
            score: The new score to assign to the track.
            upd_id: The update ID associated with the detection.
        """
        self.score = score
        self.num_det += 1
        self.num_miss = 0
        self.upd_id = upd_id

    def bump_num_misses(self) -> None:
        """
        Updates the track to reflect a missed detection.
        Reduces the score by half, resets the number of detections, and increments the number of misses.
        """
        self.score /= 2.0
        self.num_det = 0
        self.num_miss += 1
        self.upd_id = UPD_ID_LOOSE

    def __repr__(self) -> str:
        """
        Returns a string representation of the track, including the current posterior state of the Kalman filter.

        Returns:
            str: A string representation of the track.
        """
        return f'TrackNdKkf(x = {self.kkf.kalman_filter.statePost[:, 0]})'



"""Support of the block-diagonal measurement model.

The measurement model is given by a sequence of kinematic orders.
For example, if the ND state vector has one 3D part with the constant-velocity
(CV) point-motion model, then the sequence of state orders contains just one
order [2]. In this case, the sequence of measurement orders could be either [1]
or [2]. If the sequence of measurement orders is [1], then the measurements
vectors will include coordinates. If the sequence of measurement orders is [2],
then the measurements vectors will include coordinates and velocities.
"""

from typing import Generator, Sequence

import numpy as np

from .gen_x import NdKkfMatGenX


def assert_orders_z_compatible(gen_x: NdKkfMatGenX, orders_z: Sequence[int]) -> None:
    """Assert compatibility of the state-matrix generator and the measurement orders.

    Args:
        gen_x: the generator of state matrices.
        orders_z: the sequence of the measurement orders.

    Raises:
        AssertError: if number of parts are different or any of the measurement
                     orders exceeds the corresponding state order.
    """
    assert len(gen_x.orders) == len(orders_z), (
        f'Length should be equal, but {len(gen_x.orders)} /= {len(orders_z)}.'
    )
    for order_x, order_z in zip(gen_x.orders, orders_z):
        assert order_z <= order_x, f'Measurement should have <= order, but {order_z} > {order_x}'
        assert order_z > 0, f'Measurement should contain some variables, but {order_z} < 1'


class NdKkfMatGenXz:
    """Generator of the block-diagonal, n-dimensional, kinematic, Kalman-filter matrices."""

    def __init__(self, gen_x: NdKkfMatGenX, orders_z: Sequence[int]) -> None:
        """Constructor.

        This object keeps the resulting ND measurement model (H matrix).

        In Python, in many case the conversion from the measurement vector z to
        state vector x is possible and faster by matrix multiplication x = H^T z.
        However, we decide for the loops as a close programming-solution model
        for lower-level languages.

        Args:
            gen_x: the generator of state matrices.
            orders_z: the measurement orders for every part of the state vector.
        """
        assert_orders_z_compatible(gen_x, orders_z)
        self.gen_x = gen_x
        self.orders_z = np.array(orders_z, dtype=int)
        self.num_p = len(orders_z)
        self.num_z = int(np.sum(self.orders_z * gen_x.num_dof))
        self.h_mat = np.zeros((self.num_z, gen_x.num_x))
        for o_x, s_x, e_x, o_z, s_z, e_z in self.gen_loop_xz():
            self.h_mat[s_z:e_z, s_x : s_x + o_z] = np.eye(o_z)

    def fill_vec_x(self, vec_z: np.ndarray, vec_x: np.ndarray) -> None:
        """Distribute the variables from the measurement vector to the state vector.

        It is a block-by-block distribution.

        Args:
            vec_z: the measurement vector.
            vec_x: the state vector.
        """
        assert vec_z.shape == (self.num_z,)
        assert vec_x.shape == (self.gen_x.num_x,)
        for o_x1, s_x1, _, o_z1, s_z1, e_z1 in self.gen_loop_xz():
            vec_x[s_x1 : s_x1 + o_z1] = vec_z[s_z1:e_z1]

    def fill_mat_xx(self, mat_zz: np.ndarray, mat_xx: np.ndarray) -> None:
        """Distribute the measurement covariance to the process-noise (state) covariance.

        Args:
            mat_zz: measurement covariance matrix.
            mat_xx: process-noise covariance matrix to be partially filled.
        """
        assert mat_zz.shape == (self.num_z, self.num_z)
        assert mat_xx.shape == (self.gen_x.num_x, self.gen_x.num_x)
        for o_x1, s_x1, _, o_z1, s_z1, e_z1 in self.gen_loop_xz():
            for o_x2, s_x2, _, o_z2, s_z2, e_z2 in self.gen_loop_xz():
                mat_xx[s_x1 : s_x1 + o_z1, s_x2 : s_x2 + o_z2] = mat_zz[s_z1:e_z1, s_z2:e_z2]

    def gen_loop_xz(self) -> Generator[tuple[int, int, int, int, int, int], None, None]:
        """Generator of the loop over 1D blocks of the state and measurement vectors.

        Returns:
            Generator yielding a complete sequence of orders, start and end indices
            for state and measurement vectors.
        """
        s_x = 0
        s_z = 0
        for o_x, o_z, nd in zip(self.gen_x.orders, self.orders_z, self.gen_x.num_dof):
            for _iz in range(nd):
                e_x = s_x + o_x
                e_z = s_z + o_z
                yield o_x, s_x, e_x, o_z, s_z, e_z
                s_x = e_x
                s_z = e_z



"""."""

import numpy as np
import pytest

from cv2 import KalmanFilter

from kinematic_tracker.nd.gen_kf import NdKkfGenKf
from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz


def test_get_kf(gen_kf: NdKkfGenKf, gen_xz: NdKkfMatGenXz, kf: KalmanFilter) -> None:
    """."""
    assert kf.transitionMatrix == pytest.approx(np.zeros((5, 5)))
    assert id(kf.transitionMatrix) == id(gen_kf.f_mat)
    assert id(kf.measurementMatrix) == id(gen_xz.h_mat)

    assert kf.controlMatrix is None
    assert kf.measurementNoiseCov == pytest.approx(3.0 * np.eye(3))
    h_ref = [[1.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 1.0]]
    assert kf.measurementMatrix == pytest.approx(np.array(h_ref))
    assert kf.processNoiseCov == pytest.approx(np.zeros((5, 5)))
    p_ref = np.array(
        [
            [3.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 100.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 3.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 200.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 3.0],
        ]
    )
    assert kf.errorCovPre == pytest.approx(p_ref)
    assert kf.errorCovPost == pytest.approx(p_ref)
    x_ref = np.array([1.0, 0.0, 2.0, 0.0, 3.0]).reshape((5, 1))
    assert kf.statePre == pytest.approx(x_ref)
    assert kf.statePost == pytest.approx(x_ref)



"""."""

import numpy as np

from kinematic_tracker.proc_noise.nd_kkf_dia_noise import NdKkfDiaNoise


def test_nd_kkf_dia_noise_save_values(dia_n: NdKkfDiaNoise) -> None:
    """."""
    dia_n.save_values(np.empty(0))



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.match_hungarian import MatchDriverHungarian


@pytest.fixture
def driver() -> MatchDriverHungarian:
    driver = MatchDriverHungarian(4, 5)
    driver.num_matches = 4
    driver.reports[:] = 9
    driver.targets[:] = 9
    return driver


def test_high_threshold(
    metric_23: np.ndarray[tuple[2, 3], np.dtype[float]], driver: MatchDriverHungarian
) -> None:
    """."""
    driver.compute_matches(metric_23, 5.5)
    assert driver.num_matches == 1
    assert driver.reports == pytest.approx([1, 9, 9, 9])
    assert driver.targets == pytest.approx([2, 9, 9, 9])


def test_low_threshold(
    metric_23: np.ndarray[tuple[2, 3], np.dtype[float]], driver: MatchDriverHungarian
) -> None:
    """."""
    driver.compute_matches(metric_23, 0.5)
    assert driver.num_matches == 2
    assert driver.reports == pytest.approx([0, 1, 9, 9])
    assert driver.targets == pytest.approx([2, 1, 9, 9])


def test_all_zeros(driver: MatchDriverHungarian) -> None:
    """."""
    driver.compute_matches(np.zeros((2, 3)), 0.5)
    assert driver.num_matches == 0
    assert driver.reports == pytest.approx([9, 9, 9, 9])
    assert driver.targets == pytest.approx([9, 9, 9, 9])



"""."""

from pathlib import Path

from binary_classification_ratios import BinaryClassificationRatios

from kinematic_tracker.tracker.tracker import NdKkfTracker
from test.with_ext_data.conftest import AlignedCuboidDetectionsFromCsv


def test_nu_scenes_coo3_hungarian(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_association_method('hungarian')
    det_path = Path(__file__).parent / 'share/fusion-lab-detections-sensor-1.csv'
    det_gen = AlignedCuboidDetectionsFromCsv(str(det_path))
    tracking_history = det_gen.run_tracking(tracker)

    association_quality = det_gen.analyse_association(tracking_history)
    conf_mat_dct = association_quality.get_confusion_matrix()
    ratios = BinaryClassificationRatios(**conf_mat_dct)
    ratios.assert_min(0.9964, 0.998, 0.998)



"""."""

import pytest

from kinematic_tracker.tracker.track import NdKkfTrack


def test_nd_kkf_track_bump_num_detections(track: NdKkfTrack) -> None:
    """."""
    track.num_miss = 3
    track.bump_num_detections(0.5, 77)
    assert track.score == pytest.approx(0.5)
    assert track.num_det == 1
    assert track.num_miss == 0
    assert track.upd_id == 77


def test_nd_kkf_track_bump_num_misses(track: NdKkfTrack) -> None:
    """."""
    track.num_det = 3
    track.bump_num_misses()
    assert track.score == pytest.approx(0.0)
    assert track.num_det == 0
    assert track.num_miss == 1
    assert track.upd_id == -9999



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_pos.fundamental import FundMatCp


def test_get_fund_mat_cp() -> None:
    """."""
    fu_mat = FundMatCp()
    fu_mat.compute(0.1)
    assert fu_mat.f_mat == pytest.approx(np.eye(1))



"""."""

from typing import Any

from .free_id import get_free_id as get_free_id_list_int


def get_free_id(tracks: list[Any]) -> int:
    """."""
    return get_free_id_list_int([track.id for track in tracks])



"""."""

import pytest

from .conftest import KinematicMatrices


def test_const_pos() -> None:
    """."""
    gen = KinematicMatrices(1)
    assert gen.f_mat_dt.shape == (1, 1)
    assert gen.q_mat_dt.shape == (1, 1)


def test_const_vel() -> None:
    """."""
    gen = KinematicMatrices(2)
    assert gen.f_mat_dt.shape == (2, 2)
    assert gen.q_mat_dt.shape == (2, 2)


def test_const_acc() -> None:
    """."""
    gen = KinematicMatrices(3)
    assert gen.f_mat_dt.shape == (3, 3)
    assert gen.q_mat_dt.shape == (3, 3)


def test_const_jrk() -> None:
    """."""
    gen = KinematicMatrices(4)
    assert gen.f_mat_dt.shape == (4, 4)
    assert gen.q_mat_dt.shape == (4, 4)


def test_exceptions() -> None:
    """."""
    with pytest.raises(AssertionError):
        KinematicMatrices(-1)

    with pytest.raises(AssertionError):
        KinematicMatrices(0)

    with pytest.raises(AssertionError):
        KinematicMatrices(5)



"""."""

import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_set_ind_pos_size(tracker1: NdKkfTracker) -> None:
    """."""
    tracker1.set_association_metric('giou')
    assert len(tracker1.tracks) == 1
    assert tracker1.metric_driver.ind_pos_size == pytest.approx((0, 1, 2, -3, -2, -1))
    tracker1.set_ind_pos_size((0, 1, 2, 3, 4, 5))
    assert len(tracker1.tracks) == 0
    assert tracker1.metric_driver.ind_pos_size == pytest.approx((0, 1, 2, 3, 4, 5))



"""."""

import cv2

from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis


def test_empty_det(filters: list[cv2.KalmanFilter], driver: MetricMahalanobis) -> None:
    metric = driver.compute_metric([], filters)
    assert metric.shape == (0, 3)



"""."""

import numpy as np
import pytest

from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


def test_compute_dyn_factors(wn_fd: NdKkfWhiteNoiseFd) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    num_d = wn_fd.gen_x.num_d
    derivatives = np.linspace(9.0, 9.0 + num_d - 1, num=num_d)
    wn_fd.compute_dyn_factors(vec_x, derivatives)
    assert wn_fd.dynamic_factors == pytest.approx([18.0, 40.0, 66.0, 12.0, 13.0])

    wn_fd.pn_meta.mult_by_last_derivative = False
    wn_fd.compute_dyn_factors(vec_x, derivatives)
    assert wn_fd.dynamic_factors == pytest.approx([9.0, 10.0, 11.0, 12.0, 13.0])


def test_compute_dyn_factors_multiplied_by_variance_factor(wn_fd: NdKkfWhiteNoiseFd) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    num_d = wn_fd.gen_x.num_d
    derivatives = np.linspace(9.0, 9.0 + num_d - 1, num=num_d)
    wn_fd.variance_factor = 2.0
    wn_fd.compute_dyn_factors(vec_x, derivatives)
    assert wn_fd.dynamic_factors == pytest.approx([36.0, 80.0, 132.0, 24.0, 26.0])



"""Module for N-dimensional kinematic Kalman filtering.

This module provides the `NdKkf` class, which implements an N-dimensional
Kalman filter for kinematic tracking. It includes methods for prediction
and correction steps, leveraging the computation of the process-noise
covariance.
"""

import copy

from typing import Any

import numpy as np

from cv2 import KalmanFilter

from kinematic_tracker.core.kalman_cv2_copy import get_copy
from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.proc_noise.nd_kkf_proc_noise_factory import TT_PN


class NdKkf:
    """N-dimensional kinematic Kalman filter.

    This class encapsulates the functionality of an N-dimensional kinematic Kalman
    filters handling several kinds of process-noise.
    """

    def __init__(self, kalman_filter: KalmanFilter, proc_noise: TT_PN) -> None:
        """Initialize the N-dimensional Kalman filter.

        Args:
            kalman_filter: An initialized OpenCV Kalman filter instance.
            proc_noise: An initialized process-noise driver.
        """
        self.proc_noise = proc_noise
        self.kalman_filter = kalman_filter

    def __repr__(self) -> str:
        """Return a string representation of the NdKkf instance.

        Returns:
            str: A string representation of the instance, including process noise details.
        """
        return f'NdKkf({self.proc_noise})'

    def __deepcopy__(self, memo: dict[str, Any]) -> 'NdKkf':
        """Create a deep copy of the NdKkf instance.

        Args:
            memo: A dictionary of objects already copied.

        Returns:
            NdKkf: A deep copy of the current instance.
        """
        return NdKkf(get_copy(self.kalman_filter), copy.deepcopy(self.proc_noise))

    def predict(self, pre: NdKkfPrecursors) -> None:
        """Perform the prediction step of the Kalman filter.

        This step updates the state prediction using the pre-computed
        1D kinematic matrices from the precursor and applying the current
        process-noise model.

        Args:
            pre: A kinematic-matrix precursor containing 1-dimensional fundamental
                 and white-noise matrices, calculated for the current time step `dt`.

        Note:
            The fundamental matrix in the precursor should already be computed
            before calling this method.

            The white-noise matrices are not used for the diagonal process-noise
            kind, but they should be computed for the other process-noise kinds.
        """
        self.proc_noise.fill_proc_cov(
            pre, self.kalman_filter.statePost[:, 0], self.kalman_filter.processNoiseCov
        )
        self.kalman_filter.predict()

    def correct(self, vec_z: np.ndarray) -> None:
        """Perform the correction step of the Kalman filter.

        This step updates the state estimate based on the provided measurement.

        Args:
            vec_z: The observation reading (measurement vector).
        """
        self.proc_noise.save_values(self.kalman_filter.statePre[:, 0])
        self.kalman_filter.correct(vec_z)



"""This is the transition matrix for the constant-position motion model for Cartesian point."""

import numpy as np


class FundMatCp:
    """."""

    def __init__(self) -> None:
        """."""
        self.f_mat = np.ones((1, 1))

    def compute(self, _dt: float) -> None:
        """."""



"""."""

import pytest

from kinematic_tracker.association.g_iou_scores_aligned import GIoUAux


def test_get_intersection(
    giou_aux: GIoUAux,
    disjoint_xyz: GIoUAux,
    disjoint_x: GIoUAux,
    disjoint_xy: GIoUAux,
    joint_xy: GIoUAux,
) -> None:
    """."""
    assert giou_aux.get_intersection(giou_aux) == pytest.approx(120.0)
    assert giou_aux.get_intersection(joint_xy) == pytest.approx(72.0)
    assert joint_xy.get_intersection(giou_aux) == pytest.approx(72.0)
    assert giou_aux.get_intersection(disjoint_x) == pytest.approx(0.0)
    assert disjoint_x.get_intersection(giou_aux) == pytest.approx(0.0)
    assert giou_aux.get_intersection(disjoint_xy) == pytest.approx(0.0)



"""."""

import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned
from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis
from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_at_start(tracker: NdKkfTracker) -> None:
    """."""
    assert isinstance(tracker.metric_driver, MetricMahalanobis)


def test_set_association_metric_iou(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_association_metric('giou')
    assert isinstance(tracker.metric_driver, MetricGIoUAligned)
    assert tracker.association.mah_pre_factor == pytest.approx(1.0)


def test_set_association_metric_mahalanobis(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_association_metric('mahalanobis', 0.567)
    assert isinstance(tracker.metric_driver, MetricMahalanobis)
    assert tracker.association.mah_pre_factor == pytest.approx(0.567)



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_jerk.fundamental import FundMatCj


def test_get_fund_mat_cj() -> None:
    """."""
    gen = FundMatCj()
    gen.compute(12.0)
    f_mat = gen.f_mat
    ref = [1.0, 12.0, 72.0, 288.0, 0.0, 1.0, 12.0, 72.0, 0.0, 0.0, 1.0, 12.0, 0.0, 0.0, 0.0, 1.0]
    assert f_mat == pytest.approx(np.array(ref).reshape(4, 4))



"""Produce a derivative-sorted ordering of the variables.

Convert the block-diagonal ordered vectors and matrices to a derivative-sorted ordering.
For example, for an ND state with orders [3, 1] and number of dimensions [2, 2],
the block-diagonal state vector (bd) used internally, is going to have entries

   bd^T = (px, vx, ax, py, vy, ay, dx, dy)

The corresponding derivative-sorted state vector (ds) is

   ds^T = (px, py, dx, dy, vx, vy, ax, ay)

The derivative-sorted ordering might simplify input-output processing.
"""

from typing import Sequence

import numpy as np

from .gen_x import NdKkfMatGenX


class DerivativeSorted:
    """The helper for the derivative-sorted ordering of the variables."""

    def __init__(self, gen_x: NdKkfMatGenX) -> None:
        """Constructor

        Args:
            gen_x: the state-vector generator.
        """
        nx = gen_x.num_x
        self.tmp_mat = np.zeros((nx, nx))
        self.c_mat = np.zeros((nx, nx))
        self.bd_ds = np.zeros(nx, dtype=int)
        ind = 0
        for der_order in range(4):
            for part_order, s, _ in gen_x.gen_loop():
                if der_order < part_order:
                    self.bd_ds[s + der_order] = ind
                    self.c_mat[ind, s + der_order] = 1.0
                    ind += 1

    def convert_vec(self, vec_x: np.ndarray, out_x: np.ndarray) -> None:
        """Convert from block-diagonal to derivative-sorted ordering for a single vector.

        Args:
            vec_x: the block-diagonal vector.
            out_x: the buffer to fill with the corresponding derivative-sorted vector.
        """
        out_x[self.bd_ds] = vec_x

    def convert_vectors(self, vec_sx: Sequence[np.ndarray], out_sx: Sequence[np.ndarray]) -> None:
        """Convert to derivative-sorted ordering for a sequence of state vectors.

        Args:
            vec_sx: the input sequence of vectors with block-diagonal ordering of variables.
            out_sx: the buffers to fill with derivative-sorted ordering.
        """
        for vec_x, out_x in zip(vec_sx, out_sx):
            self.convert_vec(vec_x, out_x)

    def convert_mat(self, in_xx: np.ndarray, out_xx: np.ndarray) -> None:
        """Convert the block-diagonal covariance matrices to the derivative-sorted covariances.

        Args:
            in_xx: the input matrix with block-diagonal ordering of variables.
            out_xx: the output matrix with derivative-sorted ordering of variables.
        """
        np.dot(self.c_mat, in_xx, out=self.tmp_mat)
        np.dot(self.tmp_mat, self.c_mat.T, out=out_xx)

    def get_num_x(self) -> int:
        """Get the number of variables in the state vectors

        Returns:
            The number of variables in the state vectors.
        """
        return self.c_mat.shape[0]



"""."""

from pathlib import Path

import numpy as np

from association_quality_clavia import AssociationQuality
from binary_classification_ratios import BinaryClassificationRatios

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_association_quality() -> None:
    """."""

    det_path = Path(__file__).parent / 'detections.csv'
    tss_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
    ids_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
    det_lz = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=range(2, 8))
    stamps = np.unique(tss_l)
    # Detections are prepared above.

    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.set_measurement_std_dev(0, 0.05)  # define std dev for coordinates and...
    tracker.set_measurement_std_dev(1, 0.02)  # ... for dimensions of cuboids
    # Tracker is initialized, tracking will be run below.

    classifier = AssociationQuality()  # Classifier into TP, FP, FN, TN
    for stamp_ns in stamps:
        mask = tss_l == stamp_ns
        det_rz, ids_r = det_lz[mask], ids_l[mask]
        tracker.advance(stamp_ns, det_rz, ids_r)
        for track in tracker.tracks:
            is_det_supplied = track.ann_id in ids_r
            classifier.classify(track.ann_id, track.upd_id, is_det_supplied)

    conf_mat_dct = classifier.get_confusion_matrix()
    ratios = BinaryClassificationRatios(**conf_mat_dct)
    print(ratios.get_summary())
    ratios.assert_min(0.8918, 0.9939, 0.8878)



"""
This module provides functionality for calculating the Generalized Intersection over Union (GIoU)
between two 3D bounding boxes. The bounding boxes are aligned with Cartesian axes.
It includes a helper class `GIoUAux` that encapsulates the necessary operations and
intermediate quantities for computing GIoU.
"""

import numpy as np


class GIoUAux:
    """
    A helper class for calculating the Generalized Intersection over Union (GIoU)
    between two 3D bounding boxes aligned with Cartesian axes. The object represents
    one bounding box. The bounding box is defined by the coordinates of its geometrical
    center and its extent along each of the Cartesian axes.
    """

    def __init__(self) -> None:
        """
        Initializes the GIoUAux object with nil values for the bounding box
        corners, volume, and intermediate arrays used in calculations.
        """
        self.corners: np.ndarray = np.zeros(
            (2, 3)
        )  # Stores the min and max corners of the bounding box
        self.volume = 0.0  # Volume of the bounding box
        self.min_max: np.ndarray = np.zeros(3)  # Temporary array for min of max corners
        self.max_min: np.ndarray = np.zeros(3)  # Temporary array for max of min corners
        self.min_min: np.ndarray = np.zeros(3)  # Temporary array for min of min corners
        self.max_max: np.ndarray = np.zeros(3)  # Temporary array for max of max corners
        self.diff: np.ndarray = np.zeros(3)  # Temporary array for differences in dimensions

    def set_vec_z(self, vec_z: np.ndarray) -> None:
        """Set the bounding box center and dimensions.

        Args:
            vec_z: The center and dimensions of the bounding box packed into one vector.
                   vec_z = (pos_x, pos_y, pos_z, dim_x, dim_y, dim_z)
        """
        flat_z = vec_z.flatten()
        self.corners[:] = flat_z[:3]  # Set the center coordinates
        half_dim = flat_z[3:6] / 2.0  # Compute half of the dimensions
        self.corners[0] -= half_dim  # Compute the minimum corner
        self.corners[1] += half_dim  # Compute the maximum corner
        self.volume = flat_z[3:6].prod()  # Compute the volume of the bounding box

    def get_g_iou_diff(self, other: 'GIoUAux') -> float:
        """
        Computes the Generalized Intersection over Union (GIoU) difference
        between this bounding box and another.

        Args:
            other (GIoUAux): Another GIoUAux object representing a bounding box.

        Returns:
            float: The GIoU difference value.
        """
        intersection_volume = self.get_intersection(other)  # Compute intersection volume
        union_volume = self.volume + other.volume - intersection_volume  # Compute union volume
        iou = intersection_volume / union_volume  # Compute IoU
        enclosing_volume = self.get_enclosing(other)  # Compute enclosing volume
        g_iou = iou - (enclosing_volume - union_volume) / enclosing_volume  # Compute GIoU
        return g_iou

    def get_g_iou(self, other: 'GIoUAux') -> float:
        """
        Computes the Generalized Intersection over Union (GIoU) between this
        bounding box and another.

        Args:
            other (GIoUAux): Another GIoUAux object representing a bounding box.

        Returns:
            float: The GIoU value, normalized to the range [0, 1].
        """
        return (self.get_g_iou_diff(other) + 1.0) / 2.0

    def get_intersection(self, other: 'GIoUAux') -> float:
        """
        Computes the intersection volume between this bounding box and another.

        Args:
            other (GIoUAux): Another GIoUAux object representing a bounding box.

        Returns:
            float: The intersection volume. Returns 0.0 if there is no intersection.
        """
        np.fmin(self.corners[1, :], other.corners[1, :], out=self.min_max)  # Min of max corners
        np.fmax(self.corners[0, :], other.corners[0, :], out=self.max_min)  # Max of min corners
        np.subtract(self.min_max, self.max_min, out=self.diff)  # Compute the difference
        return 0.0 if np.any(self.diff < 0.0) else self.diff.prod()  # Return volume or 0.0

    def get_enclosing(self, other: 'GIoUAux') -> float:
        """
        Computes the volume of the smallest enclosing box that contains both
        this bounding box and another.

        Args:
            other (GIoUAux): Another GIoUAux object representing a bounding box.

        Returns:
            float: The enclosing volume.
        """
        np.fmin(self.corners[0, :], other.corners[0, :], out=self.min_min)  # Min of min corners
        np.fmax(self.corners[1, :], other.corners[1, :], out=self.max_max)  # Max of max corners
        np.subtract(self.max_max, self.min_min, out=self.diff)  # Compute the difference
        return self.diff.prod()  # Return the enclosing volume



"""
This module defines the `DerivativeMixer` class. The class provides functionality
to compute the derivatives (finite-differences) and compute running average of these
derivatives of a given set of variables.
"""

import numpy as np


DT_TOL = 1.0e-5  # Default tolerance for the time step


class DerivativeMixer:
    """
    A class to compute and mix derivatives of values over time, with support
    for dynamic or static mixing weights.

    Attributes:
        mix_weight: The mixing weight used for running averages.
        num_var: The number of variables being tracked.
        is_dynamic: Indicates whether the mixing weight is dynamic.
        values: The current values of the variables.
        derivatives: The computed derivatives of the variables (result).
        counter: A counter used for dynamic weight computation.
        dt_tol: The tolerance for the time step.
        last_mix_weight: The last computed mixing weight.
    """

    def __init__(
        self, mix_weight: float, num_var: int, is_dynamic: bool, dt_tol: float = DT_TOL
    ) -> None:
        """
        Initializes the `DerivativeMixer` class with the given parameters.

        Args:
            mix_weight: The nominal mixing weight for averaging of derivatives.
            num_var: The number of variables to compute.
            is_dynamic: Whether the mixing weight is dynamic.
            dt_tol: The optional tolerance for the time step. Defaults to `DT_TOL`.
        """
        self.mix_weight = mix_weight
        assert num_var > 0
        self.num_var = num_var
        self.is_dynamic = is_dynamic
        self.values: np.ndarray = np.zeros(num_var)
        self.derivatives: np.ndarray = np.zeros(num_var)
        self.counter: int = 0
        self.dt_tol = dt_tol
        self.last_mix_weight = 0.0

    def get_mixing_weight(self) -> float:
        """
        Computes the mixing weight based on the current state.

        Returns:
            float: The computed mixing weight.
        """
        if self.is_dynamic:
            w = self.counter / (self.counter + 1.0)
            w = w if w < self.mix_weight else self.mix_weight
        else:
            w = self.mix_weight
        self.counter += 1
        self.last_mix_weight = w
        return w

    def save_values(self, values: np.ndarray) -> None:
        """
        Saves the current values for future derivative computation.

        Args:
            values: The values to save.
        """
        self.values[:] = values

    def compute(self, dt: float, values: np.ndarray) -> None:
        """
        Computes the derivatives of the given values based on the time step.

        Args:
            dt: The time step for derivative computation.
            values: The new values to compute derivatives from.

        The computation blends the new derivatives with the previous ones
        using the mixing weight.
        """
        if dt > self.dt_tol:
            w = self.get_mixing_weight()
            self.derivatives[:] = w * self.derivatives[:] + (1.0 - w) * (values - self.values) / dt



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_get_track(tracker: NdKkfTracker) -> None:
    """."""
    tracker.gen_kf.f_mat[:] = np.eye(12)
    vec_z = np.linspace(1.0, 6.0, num=6)
    tracker.set_measurement_cov(0.01 * np.eye(6))
    track = tracker.get_track(vec_z, 123)
    vec_x_ref = [1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 5, 6]
    assert track.kkf.kalman_filter.statePost[:, 0] == pytest.approx(vec_x_ref)
    assert track.kkf.kalman_filter.transitionMatrix == pytest.approx(np.eye(12))
    assert track.ann_id == 123
    assert track.id == -1
    assert track.creation_id == 0



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_upd_for_loose_tracks1(tracker1: NdKkfTracker) -> None:
    """."""
    tracker1.update_for_loose_tracks({0: 0}, 1)
    assert len(tracker1.tracks) == 1
    assert tracker1.tracks[0].num_miss == 0



"""Calculation of 1D kinematic matrices for a given sets of orders.

The object stores kinematic matrix (transition- and white-noise) generators
and assists in computing, buffering and accessing the 1D parts in the state
matrix ND generators.
"""

from typing import Sequence

from kinematic_tracker.matrices.factory import KinematicMatrices, get_is_order_implemented


class NdKkfPrecursors:
    """1D precursors object to organize the computation and access to 1D kinematic matrices."""

    def __init__(self, orders: Sequence[int]) -> None:
        """Precursor matrices (1D kinematic matrices) assistant.

        Provided with a sequence of orders participating in the ND vectors,
        the assistant stores the corresponding set of kinematic orders. When executing
        the computation of the 1D matrices, only orders from this set will be calculated.

        Args:
            orders: the kinematic orders of interest.
        """
        self.orders_set = set()
        self.set_orders(orders)
        self.last_dt = 0.0
        self.kin_mat_gen = [KinematicMatrices(o) if o > 0 else None for o in range(5)]

    def set_orders(self, orders: Sequence[int]) -> None:
        """Hot-change of the orders sequence.

        Args:
            orders: the kinematic orders of interest.

        Raises:
            AssertError: if the new orders are not complying with the requirements.
        """
        assert len(orders) > 0
        assert all([get_is_order_implemented(o) for o in orders])
        self.orders_set = set(orders)
        self.last_dt = 0.0

    def compute(self, dt: float) -> None:
        """Compute the 1D kinematic matrices for a given time step.

        Args:
            dt: the time step
        """
        self.last_dt = dt
        for o in self.orders_set:
            self.kin_mat_gen[o].compute_q_mat(dt)
            self.kin_mat_gen[o].compute_f_mat(dt)



"""NdKkfTracker class implementation.

This module defines the NdKkfTracker class. The tracker is responsible for managing
and updating a set of Kalman filter-based tracks to organize the multiple-object tracking.
It provides methods for associating detections with existing tracks, creating
new tracks, and removing stale tracks.

Classes:
    NdKkfTracker: A tracker that uses kinematic Kalman filters for multiple-object tracking.
"""

from typing import Sequence

import cv2
import numpy as np

from kinematic_tracker.association.association import Association, AssociationMethod
from kinematic_tracker.association.association_metric import AssociationMetric
from kinematic_tracker.core.creation_id import CreationId
from kinematic_tracker.core.loose_items import get_loose_indices
from kinematic_tracker.core.track_free_id import get_free_id
from kinematic_tracker.nd.gen_kf import NdKkfGenKf
from kinematic_tracker.nd.shape import NdKkfShape
from kinematic_tracker.proc_noise.meta import ProcNoiseMeta
from kinematic_tracker.proc_noise.nd_kkf_proc_noise_factory import get_proc_noise

from .kkf import NdKkf
from .track import NdKkfTrack


class NdKkfTracker:
    """NdKkfTracker class to manage the multiple-object tracking using the
       N-dimensional, mixed-order, kinematic Kalman filters.

    Attributes:
        shape: auxiliary object encoding the distribution of variables in N-dimensional,
               mixed-order, block-diagonal state vectors.
        gen_xz: A manager object to cope with the block-diagonal distribution of variables.
        ini_der_vars: Initial variances for diagonal parts of the error covariances not defined
                      by the measurement covariances.
        pre: Precursor 1D matrices for Kalman filter prediction step (transition- and white-noise).
        pn_meta: Metadata (parameters) for the process-noise calculation.
        creation_id: A generator of the globally unique track IDs.
        tracks: List of tracked targets.
        gen_kf: Kalman filter generator. Contains a shared buffer for transition matrices.
        association: Metadata (parameters) for association of detection reports and targets.
        metric_driver: Metric computation driver for association.
        match_driver: Match computation driver for association.
        cov_zz: Measurement covariance matrix. Initially None, must be set before use.
        num_misses_max: Maximum number of consecutive misses before target removal.
        num_det_min: Minimum number of consecutive detections required to confirm targets.
        last_ts_ns: Timestamp of the last prediction in nanoseconds.
    """

    def __init__(
        self,
        orders_x: Sequence[int],
        num_dims: Sequence[int],
        orders_z: Sequence[int] | None = None,
    ) -> None:
        """Initialize the NdKkfTracker.

        Args:
            orders_x: Kinematic orders for each part of the ND, mixed-order states.
            num_dims: Number of dimensions for each part of the ND, mixed-order state.
            orders_z: Kinematic orders of the measurement variables. The argument is optional.
                      If not provided, it defaults to a sequence of ones of suitable length.
        """
        orders_z_seq = [1] * len(num_dims) if orders_z is None else orders_z
        self.shape = NdKkfShape(orders_x, num_dims, orders_z_seq)
        self.gen_xz = self.shape.get_mat_gen_xz()
        self.ini_der_vars = 100.0 * np.ones(self.gen_xz.gen_x.num_d)
        self.pre = self.shape.get_precursors()
        self.pn_meta = ProcNoiseMeta()
        self.creation_id = CreationId()
        self.tracks: list[NdKkfTrack] = []
        num_x = self.gen_xz.gen_x.num_x
        self.gen_kf = NdKkfGenKf(np.zeros((num_x, num_x)))
        self.association = Association(num_x, self.gen_xz.num_z)
        self.metric_driver = self.association.get_metric_driver()
        self.match_driver = self.association.get_match_driver()
        self.cov_zz: np.ndarray | None = None
        self.num_misses_max = 3
        self.num_det_min = 3
        self.last_ts_ns = 0

    def internal_check(self) -> None:
        """Perform internal consistency checks.

        Raises:
            RuntimeError: If the measurement covariance is not set or contains NaN values.
        """
        if self.cov_zz is None:
            raise RuntimeError('Forgot to set the measurement covariance?')
        if np.any(np.isnan(self.cov_zz)):
            raise RuntimeError('All standard deviations defined?')

    def set_association_method(self, method_name: str) -> None:
        """Set the association method for matching detections to tracks.

        Args:
            method_name: Name of the association method.
        """
        self.association.method = AssociationMethod(method_name)
        self.match_driver = self.association.get_match_driver()

    def set_association_metric(
        self,
        metric_name: str,
        mah_factor: float = 1.0,
        ind_pos_size: Sequence[int] = (0, 1, 2, -3, -2, -1),
    ) -> None:
        """Set the association metric for evaluating matches.

        Args:
            metric_name: name of the association metric.
            mah_factor: Mahalanobis distance pre-factor. Defaults to 1.0.
            ind_pos_size: indices for extracting positions and sizes from the measurement vectors.
                          For more details, see docstring of the method `set_ind_pos_size(...)`.
        """
        self.association.metric = AssociationMetric(metric_name)
        self.association.mah_pre_factor = mah_factor
        self.association.ind_pos_size = ind_pos_size
        self.metric_driver = self.association.get_metric_driver()

    def set_association_threshold(self, threshold: float) -> None:
        """Set the matching threshold for the association.

        The threshold is a guard against clutter detections.
        This setter is very simple. It exists to assure the library consumers.

        Args:
            threshold: new threshold value.
        """
        if not (0.0 < threshold < 1.0):
            raise ValueError(f'Threshold {threshold} seems to be out of range.')
        self.association.threshold = threshold

    def set_ind_pos_size(self, ind_pos_size: Sequence[int]) -> None:
        """Set the distribution of variables in the measurement vector.

        This is relevant for GIoU metric because for this metric we treat the
        variables in a non-uniform way. Firstly, the measurement vector
        should contain the positions of the cuboids as well as their sizes (dimensions).
        Secondly, we should not mix up positions (vector with 3 variables)
        and sizes (another vector with 3 variables). Finally, the distribution
        of variables in the measurement vector could be arbitrary, so we
        need a permutation index in general. The argument `ind_pos_size` should be
        a sequence of integers pointing to the positions and sizes of the cuboids
        in the following convention:

           (pos_x, pos_y, pos_z, size_x, size_y, size_z)

        By default, the `ind_pos_size = (0, 1, 2, -3, -2, -1)`. This corresponds
        to the positions gathered at the start of the measurement vector and
        the sizes gathered at the end of the measurement vector. This choice of default
        value should be convenient for ASAM OpenLabel measurement vectors.
        The ASAM OpenLabel measurement vectors feature the following distribution of
        variables

            z^T = (px, py, pz, alpha, beta, gamma, sx, sy, sz)

        or

            z^T = (px, py, pz, a, b, c, d, sx, sy, sz)

        In both cases, the sizes of the cuboids locate at the end of the measurement vector.

        Args:
            ind_pos_size: the permutation pointer array.
        """
        self.tracks.clear()
        self.association.ind_pos_size = ind_pos_size
        self.metric_driver = self.association.get_metric_driver()

    def set_measurement_cov(self, cov_zz: np.ndarray) -> None:
        """Set (the whole) measurement covariance matrix.

        Args:
            cov_zz: Measurement covariance matrix.

        Raises:
            ValueError: If the shape of the covariance matrix is incorrect.
        """
        shape = (self.gen_xz.num_z, self.gen_xz.num_z)
        if self.cov_zz is None:
            self.cov_zz = np.zeros(shape)
        if cov_zz.shape != shape:
            raise ValueError(f'Wrong shape of the measurement covariance. Expected {shape}')
        self.cov_zz[:, :] = cov_zz

    def set_measurement_std_dev(self, part: int, std_dev: float, der_order: int = 0) -> None:
        """Set the standard deviation for a given part of the measurement vector.

        Args:
            part: Index of the measurement part.
            std_dev: Standard deviation value.
            der_order: Derivative order. Defaults to 0.

        Raises:
            ValueError: If no element is found for the given part and derivative order.
        """
        if self.cov_zz is None:
            self.cov_zz = np.eye(self.gen_xz.num_z)
            for i in range(self.gen_xz.num_z):
                self.cov_zz[i, i] = np.nan  # to prevent incomplete initialization

        ind = 0
        variance = std_dev * std_dev
        is_set = False
        for cur_part, (o, nd) in enumerate(zip(self.gen_xz.orders_z, self.gen_xz.gen_x.num_dof)):
            for _ in range(nd):
                for cur_der_order in range(o):
                    if cur_der_order == der_order and cur_part == part:
                        self.cov_zz[ind, ind] = variance
                        is_set = True
                    ind += 1
        if not is_set:
            raise ValueError(f'No element found for part {part} and der_order {der_order}')

    def __repr__(self) -> str:
        """Return a string representation of the tracker.

        Returns:
            str: String representation of the tracker.
        """
        return (
            f'NdKkfTracker(\n'
            f'    {self.association}\n'
            f'    num_misses_max {self.num_misses_max}\n'
            f'    num_det_min {self.num_det_min})'
        )

    def correct_associated(
        self,
        reports: Sequence[int],
        targets: Sequence[int],
        metric_rt: np.ndarray,
        det_rz: Sequence[np.ndarray],
        ids_r: Sequence[int],
    ) -> None:
        """Correct tracks based on the given association matches.

        Args:
            reports: Indices of associated detection reports.
            targets: Indices of associated tracked targets.
            metric_rt: Metric values for report-target pairs.
            det_rz: Detection vectors to be used eventually for correction step.
            ids_r: Detection IDs to be used eventually for assigning the update IDs.
        """
        for r, t in zip(reports, targets):
            track = self.tracks[t]
            track.kkf.correct(det_rz[r])
            track.bump_num_detections(float(metric_rt[r, t]), ids_r[r])

    def update_for_loose_tracks(self, targets: Sequence[int], num_tracks_before: int) -> None:
        """Update the list of tracks for loose tracks (i.e., tracks without association).

        Args:
            targets: Indices of associated tracks.
            num_tracks_before: Number of tracks before association.
        """
        tracks_wo_correspondence = get_loose_indices(targets, num_tracks_before)
        for t in tracks_wo_correspondence:
            self.tracks[t].bump_num_misses()

        to_remove = [t for t in self.tracks if t.num_miss > self.num_misses_max]
        for t in to_remove:
            self.tracks.remove(t)

    def get_track(self, vec_z: np.ndarray, ann_id: int) -> NdKkfTrack:
        """Create a new track for a given measurement vector, with corresponding annotation ID.

        Args:
            vec_z: Measurement (detection) vector.
            ann_id: Annotation ID of the detection report.

        Returns:
            NdKkfTrack: Newly created track.
        """
        self.internal_check()
        kf = self.gen_kf.get_kf(self.gen_xz, vec_z, self.cov_zz, self.ini_der_vars)
        proc_noise = get_proc_noise(self.gen_xz.gen_x, self.pn_meta, kf.statePre[:, 0])
        nd_kkf = NdKkf(kf, proc_noise)
        track = NdKkfTrack(nd_kkf, ann_id)
        track.creation_id = self.creation_id.get_next_id()
        return track

    def update_for_loose_reports(
        self,
        reports: np.ndarray[int, np.dtype[int]],
        det_rz: Sequence[np.ndarray],
        ids_r: Sequence[int],
    ) -> None:
        """Update the list of tracks for loose detections (detections without association).

        Args:
            reports: Indices of associated reports.
            det_rz: Detection vectors to eventually construct the states of new targets.
            ids_r: Annotation IDs to eventually assign in the new targets.
        """
        reports_wo_correspondence = get_loose_indices(reports, len(det_rz))
        for r in reports_wo_correspondence:
            self.tracks.append(self.get_track(det_rz[r], ids_r[r]))

    def promote_clutter_eventually(self) -> None:
        """Promote clutter tracks to regular tracks if they meet the
        condition on the minimal number of consecutive associations.
        """
        for track in self.tracks:
            if track.id < 0 and track.num_det > self.num_det_min:
                track.id = get_free_id(self.tracks)

    def get_filters(self) -> list[cv2.KalmanFilter]:
        """Get the list of Kalman filters for all tracks.

        Returns:
            list[cv2.KalmanFilter]: List of Kalman filters.
        """
        return [track.kkf.kalman_filter for track in self.tracks]

    def predict_all(self, time_stamp_ns: int) -> None:
        """Predict the state of all tracks to the given timestamp.

        Args:
            time_stamp_ns: Timestamp in nanoseconds.
        """
        dt = (time_stamp_ns - self.last_ts_ns) / 1e9
        assert dt >= 0.0
        self.last_ts_ns = time_stamp_ns
        self.pre.compute(dt)
        self.gen_xz.gen_x.fill_f_mat(self.pre, self.gen_kf.f_mat)
        for track in self.tracks:
            track.kkf.predict(self.pre)

    def advance(
        self, stamp_ns: int, det_rz: list[np.ndarray] | np.ndarray, ids_r: Sequence[int]
    ) -> list[NdKkfTrack]:
        """Advance tracks to the given timestamp with new detections.

        Args:
            stamp_ns: timestamp of the detections in nanoseconds.
            det_rz: detection vectors.
            ids_r: annotation IDs.

        Returns:
            list[NdKkfTrack]: Updated list of tracks.
        """
        if isinstance(det_rz, np.ndarray):
            assert det_rz.ndim == 2
            assert det_rz.shape[1] == self.gen_xz.num_z

        assert isinstance(stamp_ns, int) or isinstance(stamp_ns, np.int64), (
            f'Time stamp (nanoseconds) should be an integer or np.int64, but got {type(stamp_ns)}'
        )
        self.predict_all(stamp_ns)
        score_rt = self.metric_driver.compute_metric(det_rz, self.get_filters())
        self.match_driver.compute_matches(score_rt, self.association.threshold)
        reports, targets = self.match_driver.get_reports_targets()
        self.correct_associated(reports, targets, score_rt, det_rz, ids_r)
        num_tracks_before = len(self.tracks)
        self.update_for_loose_reports(reports, det_rz, ids_r)
        self.promote_clutter_eventually()
        self.update_for_loose_tracks(targets, num_tracks_before)
        return self.tracks



"""."""

from kinematic_tracker.association.association import Association


def test_repr(association: Association) -> None:
    """."""
    ref = 'Association(greedy mahalanobis threshold 0.56 mah_pre_factor 3.996)'
    assert str(association) == ref



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_pos.fundamental import FundMatCp
from kinematic_tracker.matrices.const_pos.proc_cov import WhiteNoiseMatCp


def test_proc_cov_cp() -> None:
    """."""
    wn_mat = WhiteNoiseMatCp()
    wn_mat.compute(2.0)
    assert wn_mat.q_mat == pytest.approx(np.array([[2.0]]))


def test_proc_cov_add_cp() -> None:
    """."""
    wn_mat = WhiteNoiseMatCp()
    wn_mat.compute(0.1)
    q1 = wn_mat.q_mat.copy()
    wn_mat.compute(0.2)
    q2 = wn_mat.q_mat.copy()
    wn_mat.compute(0.3)
    q3 = wn_mat.q_mat.copy()
    fu_mat = FundMatCp()
    f_mat = fu_mat.f_mat
    diff = np.dot(f_mat, np.dot(q1, f_mat.T)) + q2 - q3
    assert diff == pytest.approx(0.0)



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.proc_noise.meta import ProcNoiseKind, ProcNoiseMeta
from kinematic_tracker.proc_noise.nd_kkf_dia_noise import NdKkfDiaNoise
from kinematic_tracker.proc_noise.nd_kkf_proc_noise_factory import get_proc_noise
from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


def test_get_proc_noise_white_noise_fd(gen_x: NdKkfMatGenX, pn_meta_fd: ProcNoiseMeta) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    pn = get_proc_noise(gen_x, pn_meta_fd, vec_x)
    assert isinstance(pn, NdKkfWhiteNoiseFd)


def test_get_dia_noise(gen_x: NdKkfMatGenX, pn_meta_dia: ProcNoiseMeta) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    pn = get_proc_noise(gen_x, pn_meta_dia, vec_x)
    assert isinstance(pn, NdKkfDiaNoise)


def test_exceptions(gen_x: NdKkfMatGenX, pn_meta_dia: ProcNoiseMeta) -> None:
    """."""
    pn_meta_dia.noise_kind = ProcNoiseKind.UNKNOWN
    with pytest.raises(ValueError):
        get_proc_noise(gen_x, pn_meta_dia, np.linspace(1.0, 8.0, num=8))



"""."""

import numpy as np
import pytest

from .conftest import DerivativeMixer


def test_derivative_mixer_fixed(der_mix: DerivativeMixer) -> None:
    """."""
    der_mix.is_dynamic = False
    der_mix.save_values(np.ones(3))
    der_mix.compute(1.0, 2.0 * np.ones(3))
    assert der_mix.derivatives == pytest.approx(0.3 * np.ones(3))

    der_mix.save_values(2.5 * np.ones(3))
    der_mix.compute(1.0, 3.0 * np.ones(3))
    assert der_mix.derivatives == pytest.approx(0.36 * np.ones(3))



"""Factory module for creating kinematic-matrix generators for a given point-motion model (order).

This module provides a `KinematicMatrices` class to compute and store the
white-noise covariance matrices Q and state transition (fundamental) matrices F
for different point-motion models:

   - constant position, kinematic order 1
   - constant velocity, kinematic order 2
   - constant acceleration, kinematic order 3
   - constant jerk, kinematic order 4.
"""

from .const_acc.fundamental import FundMatCa
from .const_acc.proc_cov import WhiteNoiseMatCa
from .const_jerk.fundamental import FundMatCj
from .const_jerk.proc_cov import WhiteNoiseMatCj
from .const_pos.fundamental import FundMatCp
from .const_pos.proc_cov import WhiteNoiseMatCp
from .const_vel.fundamental import FundMatCv
from .const_vel.proc_cov import WhiteNoiseMatCv


def get_is_order_implemented(order: int) -> bool:
    """
    Check if the given kinematic order is implemented.

    Args:
        order: The kinematic order to check.

    Returns:
        bool: True if the order is implemented (1 to 4), False otherwise.
    """
    return 0 < order < 5


class KinematicMatrices:
    """
    A class to generate and compute 1D kinematic matrices for different orders.

    Attributes:
        num_x: The kinematic order (1 to 4).
        q_mat_gen: The generator for the process noise covariance matrix Q.
        f_mat_gen: The generator for the state transition matrix F.
        q_mat_dt: The white-noise covariance matrix Q (result buffer) for a given time step.
        f_mat_dt: The state transition matrix F (result buffer) for a given time step.
    """

    def __init__(self, num_x: int) -> None:
        """
        Initialize the KinematicMatrices object.

        Args:
            num_x: The kinematic order (1 to 4).

        Raises:
            AssertionError: If the given kinematic order is not implemented.
        """
        assert get_is_order_implemented(num_x), f'Not implemented num_x {num_x}'
        self.num_x = num_x
        if num_x == 1:
            self.q_mat_gen = WhiteNoiseMatCp()
            self.f_mat_gen = FundMatCp()
        elif num_x == 2:
            self.q_mat_gen = WhiteNoiseMatCv()
            self.f_mat_gen = FundMatCv()
        elif num_x == 3:
            self.q_mat_gen = WhiteNoiseMatCa()
            self.f_mat_gen = FundMatCa()
        elif num_x == 4:
            self.q_mat_gen = WhiteNoiseMatCj()
            self.f_mat_gen = FundMatCj()
        self.q_mat_dt = self.q_mat_gen.q_mat
        self.f_mat_dt = self.f_mat_gen.f_mat

    def compute_q_mat(self, dt: float) -> None:
        """
        Compute the process noise covariance matrix (Q) for a given time step.

        Args:
            dt: The time step for which to compute the matrix.
        """
        self.q_mat_gen.compute(dt)

    def compute_f_mat(self, dt: float) -> None:
        """
        Compute the state transition matrix (F) for a given time step.

        Args:
            dt: The time step for which to compute the matrix.
        """
        self.f_mat_gen.compute(dt)



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_internal_check_undefined_cov_zz(tracker: NdKkfTracker) -> None:
    """."""
    with pytest.raises(RuntimeError):
        tracker.internal_check()


def test_internal_check_incomplete_cov_zz(tracker: NdKkfTracker) -> None:
    """."""
    tracker.cov_zz = np.eye(6)
    tracker.cov_zz[1, 2] = np.nan
    with pytest.raises(RuntimeError):
        tracker.internal_check()



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_jerk.fundamental import FundMatCj
from kinematic_tracker.matrices.const_jerk.proc_cov import WhiteNoiseMatCj


def test_proc_cov_cj() -> None:
    """."""
    wn_gen = WhiteNoiseMatCj()
    wn_gen.compute(252.0)
    cov = wn_gen.q_mat
    # fmt: off
    ref = [256096265048064., 3556892570112., 33875167334.4, 168031584.,
           3556892570112.,  50812751001.6,    504094752.,   2667168.,
           33875167334.4,     504094752.,      5334336.,     31752.,
           168031584.,       2667168.,        31752.,       252.]
    # fmt: on
    assert cov == pytest.approx(np.array(ref).reshape(4, 4))


def test_proc_cov_add() -> None:
    """."""
    wn_gen = WhiteNoiseMatCj()
    wn_gen.compute(0.1)
    q1 = wn_gen.q_mat.copy()
    wn_gen.compute(0.2)
    q2 = wn_gen.q_mat.copy()
    wn_gen.compute(0.3)
    q3 = wn_gen.q_mat.copy()
    fu_gen = FundMatCj()
    fu_gen.compute(0.2)
    f_mat = fu_gen.f_mat
    diff = np.dot(f_mat, np.dot(q1, f_mat.T)) + q2 - q3
    assert diff == pytest.approx(0.0)



"""."""

import pytest

from .conftest import DerivativeMixer


def test_derivative_mixer_exception_in_constructor() -> None:
    """."""
    with pytest.raises(AssertionError):
        DerivativeMixer(0.7, 0, False)

    with pytest.raises(AssertionError):
        DerivativeMixer(0.7, -1, False)



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors


def test_get_q_mat_compatible_factors(pre: NdKkfPrecursors, gen_x: NdKkfMatGenX) -> None:
    """."""
    pre.compute(3.0)
    cov = 1.3 * np.ones((8, 8))
    factors = np.linspace(1.0, 5.0, num=5)
    gen_x.fill_q_mat(pre, factors, cov)
    ref = [
        [9.0, 4.5, 1.3, 1.3, 1.3, 1.3, 1.3, 1.3],
        [4.5, 3.0, 1.3, 1.3, 1.3, 1.3, 1.3, 1.3],
        [1.3, 1.3, 18.0, 9.0, 1.3, 1.3, 1.3, 1.3],
        [1.3, 1.3, 9.0, 6.0, 1.3, 1.3, 1.3, 1.3],
        [1.3, 1.3, 1.3, 1.3, 27.0, 13.5, 1.3, 1.3],
        [1.3, 1.3, 1.3, 1.3, 13.5, 9.0, 1.3, 1.3],
        [1.3, 1.3, 1.3, 1.3, 1.3, 1.3, 12.0, 1.3],
        [1.3, 1.3, 1.3, 1.3, 1.3, 1.3, 1.3, 15.0],
    ]
    assert cov == pytest.approx(np.array(ref))


def test_get_q_mat_incompatible_precursors(gen_x: NdKkfMatGenX) -> None:
    """."""
    cov = np.empty((8, 8))
    factors = np.linspace(1.0, 5.0, num=5)
    with pytest.raises(AssertionError):
        gen_x.fill_q_mat(NdKkfPrecursors([3, 1]), factors, cov)


def test_get_q_mat_incompatible_factors(pre: NdKkfPrecursors, gen_x: NdKkfMatGenX) -> None:
    """."""
    cov = np.empty((8, 8))
    factors = np.linspace(1.0, 5.0, num=4)
    with pytest.raises(AssertionError):
        gen_x.fill_q_mat(pre, factors, cov)



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.g_iou_scores_aligned import GIoUAux


def test_fixtures(
    giou_aux: GIoUAux,
    disjoint_xyz: GIoUAux,
    disjoint_x: GIoUAux,
    disjoint_xy: GIoUAux,
    joint_xy: GIoUAux,
) -> None:
    """."""
    assert giou_aux.volume == pytest.approx(120.0)
    assert giou_aux.corners == pytest.approx(np.array([[-1.0, -0.5, 0.0], [3.0, 4.5, 6.0]]))
    assert disjoint_xyz.corners == pytest.approx(np.array([[9.0, 9.5, 10.0], [13.0, 14.5, 16.0]]))
    assert disjoint_x.corners == pytest.approx(np.array([[9.0, -0.5, 0.0], [13.0, 4.5, 6.0]]))
    assert disjoint_xy.corners == pytest.approx(np.array([[9.0, 9.5, 0.0], [13.0, 14.5, 6.0]]))
    assert joint_xy.corners == pytest.approx(np.array([[0.0, 0.5, 0.0], [4.0, 5.5, 6.0]]))



"""This is the transition matrix for the constant-velocity motion model for Cartesian point."""

import numpy as np


class FundMatCv:
    """."""

    def __init__(self) -> None:
        """."""
        self.f_mat = np.eye(2)

    def compute(self, dt: float) -> None:
        """."""
        self.f_mat[0, 1] = dt



"""."""

import cv2

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


def test_empty_det(filters: list[cv2.KalmanFilter], driver: MetricGIoUAligned) -> None:
    """."""
    driver = MetricGIoUAligned(100, 500, 6, (0, 1, 2, 3, 4, 5))
    metric = driver.compute_metric([], filters)
    assert metric.shape == (0, 3)



"""."""

from cv2 import CV_64F, KalmanFilter


def get_copy(kf_in: KalmanFilter) -> KalmanFilter:
    """."""
    nz, nx = kf_in.measurementMatrix.shape
    kf = KalmanFilter(nx, nz, 0, CV_64F)
    kf.measurementMatrix = kf_in.measurementMatrix.copy()
    kf.measurementNoiseCov = kf_in.measurementNoiseCov.copy()
    kf.transitionMatrix = kf_in.transitionMatrix.copy()
    kf.processNoiseCov = kf_in.processNoiseCov.copy()
    kf.errorCovPost = kf_in.errorCovPost.copy()
    kf.errorCovPre = kf_in.errorCovPre.copy()
    kf.statePost = kf_in.statePost.copy()
    kf.statePre = kf_in.statePre.copy()
    kf.gain = kf_in.gain.copy()
    return kf



"""."""

import numpy as np
import pytest

from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


def test_nd_kkf_proc_noise_gather_last_vars(wn_fd: NdKkfWhiteNoiseFd) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    last_vars = -999 * np.ones(5)
    wn_fd.gather_last_vars(vec_x, last_vars)
    assert last_vars == pytest.approx([2, 4, 6, 7, 8])



"""Module implementing a greedy matching algorithm for associating reports and targets."""

import numpy as np

from .match_driver_base import MatchDriverBase


class MatchDriverGreedy(MatchDriverBase):
    """Class implementing a greedy matching algorithm for associating reports and targets."""

    def __init__(self, num_rows_max: int, num_cols_max: int) -> None:
        """
        Initialize the MatchDriverGreedy instance.

        Args:
            num_rows_max: Maximum number of rows allowed in the score matrix.
            num_cols_max: Maximum number of columns allowed in the score matrix.
        """
        super().__init__(num_rows_max, num_cols_max)

    def compute_matches(self, score: np.ndarray, threshold: float) -> None:
        """
        Compute the association map using the greedy algorithm.

        Args:
            score: A 2D array of likelihoods where each element represents
                   the similarity score between a report and a target.
            threshold: The minimum score required to consider a pair as associated.

        Returns:
            None: Updates the `reports` and `targets` attributes with the associations.
        """
        self.check_max_shape(score.shape)
        self.num_matches = 0
        this_num_matches_max = min(score.shape)

        # Flatten the score matrix and sort indices in descending order of scores.
        args_sort = np.argsort(-score, axis=None)

        # Iterate over sorted indices to find valid associations.
        for arg in args_sort:
            # Convert the flattened index back to 2D indices (row, column).
            target_report = np.unravel_index(arg, score.shape)
            r, c = int(target_report[0]), int(target_report[1])

            # Check if the score meets the threshold and the row/column are not already associated.
            if (
                score[r, c] >= threshold
                and (r not in self.reports[: self.num_matches])
                and (c not in self.targets[: self.num_matches])
            ):
                self.reports[self.num_matches] = r
                self.targets[self.num_matches] = c
                self.num_matches += 1

            # Stop if the maximum number of matches is reached.
            if self.num_matches >= this_num_matches_max:
                break



"""."""

import numpy as np
import pytest

from .conftest import KinematicMatrices, get_is_order_implemented


def test_is_order_implemented() -> None:
    """."""
    assert not get_is_order_implemented(0)
    for o in range(1, 5):
        assert get_is_order_implemented(o)
    assert not get_is_order_implemented(5)


def test_kinematic_matrices_compute() -> None:
    """."""
    km = KinematicMatrices(3)
    km.compute_f_mat(8.0)
    km.compute_q_mat(2.0)
    f_mat_ref = [[1.0, 8.0, 32.0], [0.0, 1.0, 8.0], [0.0, 0.0, 1.0]]
    q_mat_ref = [
        [1.6, 2.0, 1.3333333333333333],
        [2.0, 2.6666666666666665, 2.0],
        [1.3333333333333333, 2.0, 2.0],
    ]
    assert km.f_mat_dt == pytest.approx(np.array(f_mat_ref))
    assert km.q_mat_dt == pytest.approx(np.array(q_mat_ref))



"""."""

from pathlib import Path

import numpy as np

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_cuboids_aligned_with_cartesian_axes_read_from_csv() -> None:
    """."""

    det_path = Path(__file__).parent / 'detections.csv'
    tss_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
    ids_l = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
    det_lz = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=range(2, 8))
    stamps = np.unique(tss_l)
    # Detections are prepared above.

    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.set_measurement_std_dev(0, 0.1)  # define std dev for coordinates and...
    tracker.set_measurement_std_dev(1, 0.02)  # ... for dimensions of cuboids
    # Tracker is initialized, tracking will be run below.

    print()
    np.set_printoptions(linewidth=200, precision=4)
    for stamp_ns in stamps:
        mask = tss_l == stamp_ns
        det_rz, ids_r = det_lz[mask], ids_l[mask]
        tracker.advance(stamp_ns, det_rz, ids_r)
        for track in tracker.tracks:
            print(track.kkf.kalman_filter.statePost.T)  # result of tracking
            # Block-diagonal variable order: px,vx,ax,py,vy,ay,pz,vz,az,dx,dy,dz
    # Tracking is done.



"""
This module defines the `MetricMahalanobis` class. The class extends the `MetricDriverBase` class
to compute Mahalanobis-distance similarity score.
"""

from typing import Sequence

import cv2
import numpy as np

from .det_type import DT, MT
from .inv_out import inv_out
from .metric_driver_base import MetricDriverBase


class MetricMahalanobis(MetricDriverBase):
    """
    A class for computing Mahalanobis distance-based metrics for associating detections
    targets represented by Kalman filters. It extends the `MetricDriverBase` class.

    Attributes:
        mah_pre_factor: Pre-factor for scaling the Mahalanobis distance.
        tmp_zx: Temporary rectangular matrix for intermediate calculations.
        cov_zz: Covariance matrix for the sum of the covariances.
        inv_zz: Inverse of the covariance matrix.
        num_z: Dimensionality of the measurement space (number of detection variables).
    """

    def __init__(
        self,
        num_reports_max: int,
        num_targets_max: int,
        mah_pre_factor: float,
        num_x: int,
        num_z: int,
    ) -> None:
        """
        Initializes the MetricMahalanobis instance with the given parameters.

        Args:
            num_reports_max: Maximum number of reports.
            num_targets_max: Maximum number of targets.
            mah_pre_factor: Pre-factor for scaling the (square) of Mahalanobis distance.
            num_x: Dimensionality of the state space used for allocating auxiliary buffers.
            num_z: Dimensionality of the measurement space used for allocating auxiliary buffers.

        Raises:
            AssertionError: If `num_x` is less than `num_z`.
        """
        assert num_x >= num_z, f'Mixed up x and z? num_x = {num_x} < {num_z} = num_z'
        super().__init__(num_reports_max, num_targets_max, num_z)
        self.mah_pre_factor = mah_pre_factor
        self.tmp_zx = np.zeros((num_z, num_x))
        self.cov_zz = np.zeros((num_z, num_z))
        self.inv_zz = np.zeros((num_z, num_z))
        self.num_z = num_z

    def compute_metric(self, det_rz: DT, filters: Sequence[cv2.KalmanFilter]) -> MT:
        """
        Computes the Mahalanobis distance-based metric for associating detections
        with Kalman filters.

        Args:
            det_rz: The input sequence of detection reports (measurements).
            filters: The input sequence of Kalman filters representing the tracked targets.

        Returns:
            np.ndarray[tuple[int, int], float]: A 2D array containing the computed metrics.
            The array has shape (num_reports, num_targets) and located at the beginning
            of the (larger) `self.metric_rt` buffer.

        Raises:
            np.linalg.LinAlgError: If the covariance matrix is singular and cannot be inverted.
        """
        rect_chunk = self.get_rect_chunk(len(det_rz), len(filters))
        for t, kf in enumerate(filters):
            np.dot(kf.measurementMatrix, kf.statePre, out=self.vec_z)
            np.dot(kf.measurementMatrix, kf.errorCovPre, out=self.tmp_zx)
            np.dot(self.tmp_zx, kf.measurementMatrix.T, out=self.cov_zz)
            np.add(self.cov_zz, kf.measurementNoiseCov, out=self.cov_zz)
            # detection does not have its noise covariance...

            inv_out(self.cov_zz, self.inv_zz)

            for r, det_z in enumerate(det_rz):
                sqr_distance = cv2.Mahalanobis(self.vec_z, det_z, self.inv_zz) ** 2
                rect_chunk[r, t] = sqr_distance
        rect_chunk *= self.mah_pre_factor / (2.0 * self.num_z)
        rect_chunk[rect_chunk > 150.0] = 150
        np.exp(-rect_chunk, out=rect_chunk)
        return rect_chunk



"""
This module defines the class WhiteNoiseMatCv, which computes and keeps a white-noise
covariance matrix for a constant velocity model. The covariance matrix is useful as
covariance of the process noise, typically after scaling by a variance prefactor.
"""

import numpy as np


class WhiteNoiseMatCv:
    """
    A class to compute the white-noise covariance matrix.

    Attributes:
        q_mat: A 2x2 matrix initialized with zeros used as result buffer.
    """

    def __init__(self) -> None:
        """Initializes the class by creating a 2x2 zeros buffer."""
        self.q_mat = np.zeros((2, 2))

    def compute(self, dt: float) -> None:
        """Computes the white-noise covariance for the given time step.

        Args:
            dt: The time step used to compute the covariance matrix.
        """
        dt2 = dt * dt
        q_mat = self.q_mat
        q_mat[0, 0] = dt2 / 3
        q_mat[0, 1] = dt / 2
        q_mat[1, 0] = q_mat[0, 1]
        q_mat[1, 1] = 1.0
        q_mat *= dt



"""."""

from kinematic_tracker.core.loose_items import get_loose_indices


def test_get_loose_values() -> None:
    """."""
    assert {0, 2, 3} == get_loose_indices([1, 4], 5)


def test_empty_correspondence() -> None:
    """."""
    assert get_loose_indices([], 8) == {0, 1, 2, 3, 4, 5, 6, 7}


def test_nothing_lost() -> None:
    """."""
    assert get_loose_indices([0, 1], 2) == set()



"""."""

import pytest

from kinematic_tracker.association.association import Association
from kinematic_tracker.association.match_greedy import MatchDriverGreedy
from kinematic_tracker.association.match_hungarian import MatchDriverHungarian


def test_get_match_driver(association: Association) -> None:
    """."""
    assert isinstance(association.get_match_driver(), MatchDriverGreedy)
    association.method = association.method.HUNGARIAN
    assert isinstance(association.get_match_driver(), MatchDriverHungarian)
    association.method = association.method.UNKNOWN_METHOD
    with pytest.raises(ValueError):
        association.get_match_driver()



"""."""

import pytest

from kinematic_tracker.nd.precursors import NdKkfPrecursors


def test_nd_kkf_pre_set_orders(pre: NdKkfPrecursors) -> None:
    """."""
    pre.set_orders([3, 4, 1])
    assert pre.orders_set == {1, 3, 4}


def test_nd_kkf_pre_set_orders_exceptions(pre: NdKkfPrecursors) -> None:
    """."""
    with pytest.raises(AssertionError):
        pre.set_orders([0, 4, 1])

    with pytest.raises(AssertionError):
        pre.set_orders([5, 1])

    with pytest.raises(AssertionError):
        pre.set_orders([])



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz


REF = [1, 404, 2, 404, 3, 404, 4, 404, 5, 6, 404, 7, 8, 404, 9, 10, 404, 11, 12, 404, 13, 14, 404]


def test_nd_kkf_mat_gen_xz_fill_vec_x(gen_xz: NdKkfMatGenXz) -> None:
    """."""
    vec_z = np.linspace(1, 14, num=14)
    vec_x = 404 * np.ones(gen_xz.gen_x.num_x)
    gen_xz.fill_vec_x(vec_z, vec_x)
    assert vec_x == pytest.approx(REF)


def test_easy_fill_vec_x(gen_xz: NdKkfMatGenXz) -> None:
    """."""
    vec_z = np.linspace(1, 14, num=14)
    vec_x = np.dot(gen_xz.h_mat.T, vec_z)
    ref_x = np.array(REF)
    ref_x[ref_x > 403] = 0.0
    assert vec_x == pytest.approx(ref_x)



"""
This module defines the `Association` class. The `Association` class holds
parameters to for computing an association metric (scores) and for computing
the bipartite matching between (detection) reports and (tracked) targets.
Apart from the parameter storage, the class provides methods to generate
the metric and match drivers based on the selected types of metric and
matching method.
"""

from typing import Sequence

from .association_method import AssociationMethod
from .association_metric import AssociationMetric
from .match_greedy import MatchDriverGreedy
from .match_hungarian import MatchDriverHungarian
from .metric_giou_aligned import MetricGIoUAligned
from .metric_mahalanobis import MetricMahalanobis
from .metric_size_modulated_mahalanobis import MetricSizeModulatedMahalanobis


MDT = MetricMahalanobis | MetricGIoUAligned | MetricSizeModulatedMahalanobis


class Association:
    """
    Represents an association mechanism for matching reports and targets
    using configurable metrics and methods. Generates metric and match drivers.
    """

    def __init__(
        self, num_x: int, num_z: int, ind_pos_size: Sequence[int] = (0, 1, 2, -3, -2, -1)
    ) -> None:
        """
        Initializes the `Association` object with default parameters.

        Args:
            num_x: Number of variables in the state vector.
            num_z: Number of variables in the measurement vector.
            ind_pos_size: indices to get positions and sizes of cuboids for GIoU.
                          By default, the indices will be 0,1,2,-3,-2,-1, i.e.
                          the center of the cuboid is taken as first three variables
                          of the measurement vector and sizes (dimensions) of the
                          cuboids are taken from the last three variables.
        """
        self.threshold = 0.25  # Threshold for association
        self.mah_pre_factor = 1.0  # Pre-factor for Mahalanobis distance
        self.method = AssociationMethod.HUNGARIAN  # Default matching method
        self.metric = AssociationMetric.GIOU  # Default metric
        self.num_reports_max = 100  # Maximum number of reports
        self.num_targets_max = 500  # Maximum number of targets
        self.num_x = num_x  # State vector dimensionality
        self.num_z = num_z  # Measurement vector dimensionality
        self.ind_pos_size = ind_pos_size  # Distribution of variables in the measurement vector.

    def __repr__(self) -> str:
        """
        Returns a string representation of the `Association` object.

        Returns:
            str: A string describing the association configuration.
        """
        return (
            f'Association({self.method.value} '
            f'{self.metric.value} '
            f'threshold {self.threshold} '
            f'mah_pre_factor {self.mah_pre_factor})'
        )

    def get_metric_driver(self) -> MDT:
        """
        Initializes and returns the metric driver based on the selected metric type.

        Returns:
            MetricMahalanobis | MetricGIoUAligned: The metric driver instance.

        Raises:
            ValueError: If the metric is not supported.
        """
        if self.metric == AssociationMetric.MAHALANOBIS:
            return MetricMahalanobis(
                self.num_reports_max,
                self.num_targets_max,
                self.mah_pre_factor,
                self.num_x,
                self.num_z,
            )
        elif self.metric == AssociationMetric.GIOU:
            return MetricGIoUAligned(
                self.num_reports_max, self.num_targets_max, self.num_z, self.ind_pos_size
            )
        elif self.metric == AssociationMetric.SIZE_MODULATED_MAHALANOBIS:
            return MetricSizeModulatedMahalanobis(
                self.num_reports_max,
                self.num_targets_max,
                self.mah_pre_factor,
                self.num_x,
                self.num_z,
                self.ind_pos_size,
            )

        else:
            raise ValueError('Unsupported metric.')

    def get_match_driver(self) -> MatchDriverGreedy | MatchDriverHungarian:
        """
        Initializes and returns the match driver based on the selected matching method.

        Returns:
            MatchDriverGreedy | MatchDriverHungarian: The match driver instance.

        Raises:
            ValueError: If the method is not supported.
        """
        method = self.method
        if method == AssociationMethod.GREEDY:
            return MatchDriverGreedy(self.num_reports_max, self.num_targets_max)
        elif method == AssociationMethod.HUNGARIAN:
            return MatchDriverHungarian(self.num_reports_max, self.num_targets_max)
        else:
            raise ValueError('Unsupported matching method.')



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_get_filters(tracker1: NdKkfTracker) -> None:
    """."""
    filters = tracker1.get_filters()
    assert len(filters) == 1



"""."""

from kinematic_tracker.tracker.track import NdKkfTrack


def test_nd_kkf_track_init(track: NdKkfTrack) -> None:
    """."""
    assert track.creation_id == -1
    assert track.ann_id == 123
    assert track.upd_id == 123
    assert track.id == -1


def test_nd_kkf_track_repr(track: NdKkfTrack) -> None:
    """."""
    assert repr(track) == 'TrackNdKkf(x = [1. 0. 2. 0. 3. 0. 4. 5.])'



"""."""

import copy

from typing import Any, Iterator

import numpy as np

from association_quality_clavia import AssociationQuality

from kinematic_tracker.nd.derivative_sorted import DerivativeSorted
from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.tracker.track import NdKkfTrack
from kinematic_tracker.tracker.tracker import NdKkfTracker


class AlignedCuboidDetectionsFromCsv:
    """."""

    def __init__(self, file_name: str) -> None:
        """."""
        self.file_name = file_name
        self.tss_s = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
        self.ids_s = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
        self.det_sz = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=range(2, 8))
        self.stamps = np.unique(self.tss_s)

    def gen_stamps_ids_vec_z(
        self,
    ) -> Iterator[tuple[int, np.ndarray[Any, float], np.ndarray[Any, int]]]:
        """."""
        for stamp in self.stamps:
            mask = self.tss_s == stamp
            det_rz, ids_r = self.det_sz[mask], self.ids_s[mask]
            yield stamp, det_rz, ids_r

    def run_tracking(self, tracker: NdKkfTracker) -> list[tuple[int, list[NdKkfTrack]]]:
        """."""
        history: list[tuple[int, list[NdKkfTrack]]] = []
        for stamp, det_rz, ids_r in self.gen_stamps_ids_vec_z():
            tracker.advance(int(stamp), det_rz, ids_r)
            assert len(tracker.tracks) >= len(ids_r)
            history.append((stamp, copy.deepcopy(tracker.tracks)))
        return history

    def get_det_ids(self, stamp: int) -> np.ndarray[Any, int]:
        """."""
        return self.ids_s[self.tss_s == stamp]

    def analyse_association(
        self, history: list[tuple[int, list[NdKkfTrack]]]
    ) -> AssociationQuality:
        """."""
        association_quality = AssociationQuality()
        for stamp, tracks in history:
            det_ids = self.get_det_ids(stamp)
            for track in tracks:
                is_supplied = track.ann_id in det_ids
                association_quality.classify(track.ann_id, track.upd_id, is_supplied)
        return association_quality


class AlignedCuboidMotionFromCsv:
    """."""

    def __init__(self, file_name: str) -> None:
        """."""
        self.file_name = file_name
        self.tss_s = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
        self.ids_s = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
        self.ref_sx = np.genfromtxt(file_name, skip_header=1, delimiter=',', usecols=range(2, 14))
        self.stamps = np.unique(self.tss_s)

    def compute_ate(self, history: list[tuple[int, list[NdKkfTrack]]]) -> float:
        """."""
        gen_x = NdKkfMatGenX([3, 1], [3, 3])
        der_sorted = DerivativeSorted(gen_x)
        post_x = np.zeros(12)
        ref_ix = []
        post_ix = []
        for stamp, tracks in history:
            mask_s = self.tss_s == stamp
            ids_r = self.ids_s[mask_s]
            ref_rx = self.ref_sx[mask_s]
            for track in tracks:
                mask_r = ids_r == track.ann_id
                ref_x = ref_rx[mask_r].reshape(12)
                der_sorted.convert_vec(track.kkf.kalman_filter.statePost.reshape(12), post_x)
                ref_ix.append(ref_x)
                post_ix.append(post_x.copy())
        ref_ix = np.array(ref_ix)
        post_ix = np.array(post_ix)
        ate = np.mean(np.linalg.norm(ref_ix[:, :3] - post_ix[:, :3], axis=1))
        return ate



"""The module for computing Generalized Intersection over Union (GIoU) association metric.

This module provides functionality for computing the similarity score between detection reports
and Kalman filter states (targets) using the Generalized Intersection over Union (GIoU) metric.
The metric is computed between the 3D bounding boxes which are aligned with Cartesian axes.

The resulting rectangular matrix will have a shape of (num_reports, num_targets), and located
at the beginning of the pre-allocated buffer `metric_rt` of the base class `MetricDriverBase`.
The resulting matrix is contiguous in memory.

Classes:
    MetricGIoUAligned: Computes GIoU-based association metrics between detections
        and Kalman filter states in aligned coordinate space.
"""

from typing import Sequence

import cv2
import numpy as np

from .det_type import DT, MT
from .g_iou_scores_aligned import GIoUAux
from .metric_driver_base import MetricDriverBase


class MetricGIoUAligned(MetricDriverBase):
    """Computes GIoU-based association metrics between detections and Kalman filter states.

    This class inherits from MetricDriverBase and implements GIoU-based metric computation
    for data association between detection reports and Kalman filter states. The computation
    is performed between cuboids (3D bounding boxes) aligned with Cartesian axes.

    Attributes:
        report_aux (GIoUAux): Helper object for processing detection measurements.
        target_aux (GIoUAux): Helper object for processing Kalman filter states.

    Args:
        num_reports_max: Maximum number of detection reports to handle.
        num_targets_max: Maximum number of targets (Kalman filters) to handle.
        num_z: Dimension of the measurement vector.
    """

    def __init__(
        self, num_reports_max: int, num_targets_max: int, num_z: int, ind_pos_size: Sequence[int]
    ) -> None:
        """Initialize the MetricGIoUAligned instance.

        Args:
            num_reports_max: maximum number of detection reports to handle.
            num_targets_max: maximum number of targets (Kalman filters) to handle.
            num_z: number of variables in the measurement vector.
            ind_pos_size: location of variables in the measurement vector.
                          By default, the indices should be 0,1,2,-3,-2,-1, i.e.
                          the center of the cuboid is taken as first three variables
                          of the measurement vector and sizes (dimensions) of the
                          cuboids are taken from the last three variables.
        """
        super().__init__(num_reports_max, num_targets_max, num_z)
        self.ind_pos_size = np.array(ind_pos_size, dtype=int)
        assert self.ind_pos_size.ndim == 1, 'Indices should be one-dimensional.'
        assert self.ind_pos_size.size == 6, 'There should be 6 fancy indices.'
        self.vec_pos_size_filter = np.zeros((6, 1))
        self.vec_pos_size_det = np.zeros((6, 1))
        self.report_aux = GIoUAux()
        self.target_aux = GIoUAux()

    def compute_metric(self, det_rz: DT, filters: Sequence[cv2.KalmanFilter]) -> MT:
        """Compute GIoU-based metrics between detections and Kalman filter states.

        Args:
            det_rz: Detection measurements, either as a sequence of numpy arrays
                or a single numpy array containing measurement vectors.
            filters: Sequence of OpenCV KalmanFilter objects representing tracked targets.

        Returns:
            np.ndarray: A matrix of GIoU scores between each detection-target pair,
                with shape (num_detections, num_targets).
        """
        rect_chunk = self.get_rect_chunk(len(det_rz), len(filters))
        for t, kf in enumerate(filters):
            np.dot(kf.measurementMatrix, kf.statePre, out=self.vec_z)
            self.vec_pos_size_filter[:, 0] = self.vec_z[self.ind_pos_size, 0]
            self.target_aux.set_vec_z(self.vec_pos_size_filter)
            for r, det_z in enumerate(det_rz):
                self.vec_pos_size_det[:, 0] = det_z[self.ind_pos_size]
                self.report_aux.set_vec_z(self.vec_pos_size_det)
                rect_chunk[r, t] = self.target_aux.get_g_iou(self.report_aux)
        return rect_chunk



"""
This module defines the `MetricDriverBase` class. The class provides a common functionality
for managing the computation of association metric (similarity scores) between
detection reports and tracked targets.
"""

import numpy as np

from .det_type import MT


class MetricDriverBase:
    """
    A base class for handling metrics between reports and targets. It provides methods
    to initialize and retrieve rectangular contiguous chunks of similarity scores.

    Attributes:
        num_reports_max: Maximum number of reports allowed.
        num_targets_max: Maximum number of targets allowed.
        num_reports: Current number of reports.
        num_targets: Current number of targets.
        num_elements: Total number of elements in the metric matrix.
        metric_rt: A 2D buffer for the similarity scores.
        vec_z: A vertical vector used as buffer in GIoU and Mahalanobis.
    """

    def __init__(self, num_reports_max: int, num_targets_max: int, num_z: int) -> None:
        """
        Initializes the MetricDriverBase instance with the given maximum limits
        for reports, targets, and additional data.

        Args:
            num_reports_max: Maximum number of reports.
            num_targets_max: Maximum number of targets.
            num_z: Number of variables in the measurement vectors.
        """
        self.num_reports_max = num_reports_max
        self.num_targets_max = num_targets_max
        self.num_reports = 0
        self.num_targets = 0
        self.num_elements = 0
        self.metric_rt = np.zeros((num_reports_max, num_targets_max))
        self.vec_z = np.zeros((num_z, 1))

    def get_rect_chunk(self, num_r: int, num_t: int) -> MT:
        """
        Retrieves a rectangular contiguous chunk of the metric matrix located in
        the buffer are carved precisely to store the similarity score between
        a given number of reports and targets.

        Args:
            num_r: number of detection reports to include in the chunk.
            num_t: number of tracked targets to include in the chunk.

        Returns:
            np.ndarray: A 2D array representing the requested rectangular chunk.

        Raises:
            ValueError: If the number of reports exceeds `num_reports_max`.
            ValueError: If the number of targets exceeds `num_targets_max`.
        """
        self.num_reports = num_r
        self.num_targets = num_t
        self.num_elements = num_r * num_t
        if num_r > self.num_reports_max:
            raise ValueError(f'Too much reports? {num_r} {self.num_reports_max}')

        if num_t > self.num_targets_max:
            raise ValueError(f'Too much targets? {num_t} {self.num_targets_max}')
        flat_chunk = self.metric_rt.reshape(-1)[: self.num_elements]
        rect_chunk = flat_chunk.reshape(num_r, num_t)
        return rect_chunk



"""."""

from kinematic_tracker.association.association_metric import AssociationMetric
from kinematic_tracker.association.metric_size_modulated_mahalanobis import (
    MetricSizeModulatedMahalanobis,
)
from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_switch_metric() -> None:
    """."""
    tracker = NdKkfTracker([2, 2], [2, 2])
    name = AssociationMetric.SIZE_MODULATED_MAHALANOBIS.value
    tracker.set_association_metric(name, ind_pos_size=(0, 1, 2, 3))
    assert isinstance(tracker.metric_driver, MetricSizeModulatedMahalanobis)



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_driver_base import MetricDriverBase


def test_metric_driver_base_init() -> None:
    """."""
    driver = MetricDriverBase(10, 50, 6)
    assert driver.metric_rt.shape == (10, 50)
    assert driver.vec_z.shape == (6, 1)


def test_metric_driver_base_get_rect_chunk() -> None:
    """."""
    driver = MetricDriverBase(4, 5, 6)
    rect_chunk = driver.get_rect_chunk(2, 3)
    assert rect_chunk.shape == (2, 3)
    assert np.shares_memory(rect_chunk, driver.metric_rt)
    driver.metric_rt[:] = 9
    rect_chunk[:] = np.linspace(1.0, 6.0, num=6).reshape((2, 3))
    # fmt: off
    ref = [1., 2., 3., 4., 5., 6., 9., 9., 9., 9., 9., 9., 9., 9., 9., 9., 9., 9., 9., 9.]
    # fmt: on
    assert ref == pytest.approx(driver.metric_rt.flatten())


def test_get_rect_chunk_exceptions() -> None:
    """."""
    driver = MetricDriverBase(4, 5, 6)
    with pytest.raises(ValueError):
        driver.get_rect_chunk(5, 3)

    with pytest.raises(ValueError):
        driver.get_rect_chunk(3, 6)



"""
This module defines the `ProcNoiseKind` enumeration. The enumerable
represents different types of process noise used in tracking.
"""

from enum import Enum


class ProcNoiseKind(Enum):
    """
    An enumeration to represent various kinds of process noise.

    Attributes:
        FINITE_DIFF: Represents the white-noise-based adaptive process noise.
        CONST: Represents the white-noise-based fixed-variance process noise.
        DIAGONAL: Represents diagonal process noise (with fixed variance).
        UNKNOWN: Represents an unknown type of process noise.
    """

    FINITE_DIFF = 'finite_diff'
    CONST = 'const'
    DIAGONAL = 'diagonal'
    UNKNOWN = 'unknown'



"""Generator of the Kalman filters.

The object facilitates generation of Kalman filters with common (shared)
transition. Sharing of the fundamental matrix across a set of Kalman filters
should improve the performance when filling the matrices. Simultaneously,
sharing of fundamental matrices reduces the memory footprint.
"""

import numpy as np

from cv2 import CV_64F, KalmanFilter

from .gen_xz import NdKkfMatGenXz


class NdKkfGenKf:
    """Generator of the Kalman filters."""

    def __init__(self, f_mat: np.ndarray) -> None:
        """Constructor.

        The object will keep the fundamental (transition) matrix and initialize the Kalman filters
        using this fundamental matrix. The fundamental matrices in Kalman filters produced by this
        generator *share the same memory*.

        Similarly, the measurement matrix (H matrix) will be taken from the measurement
        matrices generator and will share the same memory.

        Args:
            f_mat: the fundamental (transition) matrix.
        """
        self.f_mat = f_mat

    def get_kf(
        self, gen_xz: NdKkfMatGenXz, vec_z: np.ndarray, cov_zz: np.ndarray, ini_der_vars: np.ndarray
    ) -> KalmanFilter:
        """Get new Kalman filter object.

        Args:
            gen_xz: generator of Kalman matrices.
            vec_z: the measurement vector used to (partially) initialize the state vector x.
                   The derivatives of the state vector missing in the measurement will be zeros.
            cov_zz: the measurement covariance matrix used to (partially) initialize
                    the state error covariance (P matrix). The derivative parts of the
                    error covariance will be diagonal initialized with derivative variances
                    (see `ini_der_vars`)
            ini_der_vars:
                    The variances for all-order derivative parts in the initial error
                    covariances.

        For example, given a composed ND state with orders_x = [2, 1], dimensions [2, 2]
        and measurement orders orders_z = [1, 1], the state vector x will be 6-dimensional:

           x^T = (px, vx, py, vy, dx, dy)

        while the measurement vector z will be 4-dimensional

           z^T = (px, py, dx, dy)

        The given the measurement vector z^T = (px, py, dx, dy), the Kalman filter will be
        initialized with (initial) state vector x^T = (px, 0, py, 0, dx, dy).

        Similarly, given a measurement covariance

                 pp_xx  pp_xy  pd_xx  pd_xy
           R =   pp_xy  pp_yy  pd_yx  pd_yy
                 pd_xx  pd_yx  dd_xx  dd_xy
                 pd_xy  pd_yy  dd_xy  dd_yy

        and the initial derivative variances

            v^T  = (100, 200, 300, 400)

        The initial state error covariance will be

                  pp_xx  0      pp_xy  0      pd_xx  pd_xy
                  0      100    0      0      0      0
           P =    pp_xy  0      pp_yy  0      pd_yx  pd_yy
                  0      0      0      200    0      0
                  pd_xx  0      pd_yx  0      dd_xx  dd_xy
                  pd_xy  0      pd_yy  0      dd_xy  dd_yy
        """
        assert ini_der_vars.shape == (gen_xz.gen_x.num_d,)
        nx = gen_xz.gen_x.num_x
        kf = KalmanFilter(nx, gen_xz.num_z, 0, CV_64F)
        kf.measurementNoiseCov = cov_zz.copy()
        kf.measurementMatrix = gen_xz.h_mat
        kf.processNoiseCov = np.zeros((nx, nx))
        vec_x = np.zeros((nx, 1))
        gen_xz.fill_vec_x(vec_z, vec_x[:, 0])
        kf.statePre = vec_x
        kf.statePost = vec_x.copy()
        kf.errorCovPre = np.zeros((nx, nx))
        for (o, s, e), ini_der_var in zip(gen_xz.gen_x.gen_loop(), ini_der_vars):
            kf.errorCovPre[s:e, s:e] = ini_der_var * np.eye(o)
        gen_xz.fill_mat_xx(cov_zz, kf.errorCovPre)
        kf.errorCovPost = kf.errorCovPre.copy()
        kf.transitionMatrix = self.f_mat
        return kf



"""."""

import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_init_no_measurement_orders() -> None:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3])
    assert tracker.cov_zz is None
    assert tracker.gen_xz.orders_z == pytest.approx([1, 1])
    assert tracker.gen_xz.gen_x.num_x == 12
    assert tracker.gen_xz.num_z == 6


def test_nd_kkf_tracker_init_with_measurement_orders() -> None:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3], [2, 1])
    assert tracker.cov_zz is None
    assert tracker.gen_xz.orders_z == pytest.approx([2, 1])
    assert tracker.gen_xz.gen_x.num_x == 12
    assert tracker.gen_xz.num_z == 9


def test_nd_kkf_tracker_repr(tracker: NdKkfTracker) -> None:
    """."""
    ref = """NdKkfTracker(
    Association(greedy mahalanobis threshold 0.56 mah_pre_factor 1.32)
    num_misses_max 3
    num_det_min 2)"""
    assert repr(tracker) == ref



"""."""

from pathlib import Path

import numpy as np
import pytest

from kinematic_tracker.nd.derivative_sorted import DerivativeSorted
from kinematic_tracker.tracker.tracker import NdKkfTracker


def assert_quality(tracker: NdKkfTracker, ate_max: float) -> None:
    """."""
    det_path = Path(__file__).parent / 'share/fusion-lab-detections-sensor-1.csv'
    tss_s = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(0,), dtype=int)
    det_ids = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=(1,), dtype=int)
    det_sz = np.genfromtxt(det_path, skip_header=1, delimiter=',', usecols=range(2, 8))
    tracking_history = []
    for stamp, det_id, det_z in zip(tss_s, det_ids, det_sz):
        tracker.advance(int(stamp), [det_z], [det_id])
        assert len(tracker.tracks) == 1
        tracking_history.append(tracker.tracks[0].kkf.kalman_filter.statePost.reshape(12).copy())

    der_sorted = DerivativeSorted(tracker.gen_xz.gen_x)
    state_sx = np.zeros((30, 12))
    der_sorted.convert_vectors(tracking_history, state_sx)

    motion_path = det_path.parent / 'fusion-lab-motion-sensor-1.csv'
    ref_sx = np.genfromtxt(motion_path, skip_header=1, delimiter=',', usecols=range(2, 14))
    ate = np.mean(np.linalg.norm(state_sx[:, :3] - ref_sx[:, :3], axis=1))
    assert ate < ate_max


@pytest.fixture
def tracker() -> NdKkfTracker:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.set_measurement_std_dev(0, 0.5)
    tracker.set_measurement_std_dev(1, 0.02)
    return tracker



"""."""

from pathlib import Path

from binary_classification_ratios import BinaryClassificationRatios

from kinematic_tracker.tracker.tracker import NdKkfTracker
from test.with_ext_data.conftest import AlignedCuboidDetectionsFromCsv, AlignedCuboidMotionFromCsv


def test_almost_collision_wrong_std_dev(tracker: NdKkfTracker) -> None:
    """Track two almost colliding objects with wrong measurement covariance."""
    tracker.set_measurement_std_dev(0, 0.1)
    det_path = Path(__file__).parent / 'share/fusion-lab-detections-sensor-1.csv'
    det_gen = AlignedCuboidDetectionsFromCsv(str(det_path))
    tracking_history = det_gen.run_tracking(tracker)

    association_quality = det_gen.analyse_association(tracking_history)
    conf_mat_dct = association_quality.get_confusion_matrix()
    ratios = BinaryClassificationRatios(**conf_mat_dct)
    ratios.assert_min(0.408, 0.999, 0.408)

    motion_path = Path(__file__).parent / 'share/fusion-lab-motion-sensor-1.csv'
    ref_aux = AlignedCuboidMotionFromCsv(str(motion_path))
    assert ref_aux.compute_ate(tracking_history) < 8.779



"""."""

import pytest

from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.proc_noise.nd_kkf_dia_noise import NdKkfDiaNoise


@pytest.fixture
def pre() -> NdKkfPrecursors:
    return NdKkfPrecursors([2, 1])


@pytest.fixture
def dia_n() -> NdKkfDiaNoise:
    """."""
    return NdKkfDiaNoise(5.678)



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_acc.fundamental import FundMatCa
from kinematic_tracker.matrices.const_acc.proc_cov import WhiteNoiseMatCa


def test_proc_cov_ca() -> None:
    """."""
    wn_gen = WhiteNoiseMatCa()
    wn_gen.compute(30.0)
    cov = wn_gen.q_mat
    ref = np.array(
        [(1215000.0, 101250.0, 4500.0), (101250.0, 9000.0, 450.0), (4500.0, 450.0, 30.0)]
    )
    assert cov == pytest.approx(np.array(ref))


def test_proc_cov_add() -> None:
    """."""
    wn_gen = WhiteNoiseMatCa()
    wn_gen.compute(0.1)
    q1 = wn_gen.q_mat.copy()
    wn_gen.compute(0.2)
    q2 = wn_gen.q_mat.copy()
    wn_gen.compute(0.3)
    q3 = wn_gen.q_mat.copy()
    fu_gen = FundMatCa()
    fu_gen.compute(0.2)
    f_mat = fu_gen.f_mat
    diff = np.dot(f_mat, np.dot(q1, f_mat.T)) + q2 - q3
    assert diff == pytest.approx(0.0)



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker

from .conftest import assert_quality


def test_static_object_diagonal(tracker: NdKkfTracker) -> None:
    """."""
    tracker.pn_meta.noise_kind = tracker.pn_meta.noise_kind.DIAGONAL
    tracker.pn_meta.factor = 0.25
    assert_quality(tracker, 0.0130804)  # CTL shows 0.013080343950653894


def test_static_object_fini_diff(tracker: NdKkfTracker) -> None:
    """."""
    assert_quality(tracker, 0.010221)  # In CTL ATE is 0.010220974499285867



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.association import Association, AssociationMethod
from kinematic_tracker.association.association_metric import AssociationMetric


@pytest.fixture
def det_rz() -> list[np.ndarray]:
    """."""
    return [np.linspace(1.0, 6.0, num=6), np.linspace(2.0, 7.0, num=6)]


@pytest.fixture
def association() -> Association:
    """."""
    a = Association(12, 6)
    a.threshold = 0.56
    a.mah_pre_factor = 3.996
    a.method = AssociationMethod.GREEDY
    a.metric = AssociationMetric.MAHALANOBIS
    return a


@pytest.fixture
def metric_23() -> np.ndarray[tuple[2, 3], np.dtype[float]]:
    metric_rc = np.zeros((2, 3))
    # fmt: off
    metric_rc[:] = np.array(((1.0, 2.0, 4.0),
                             (4.0, 5.0, 6.0)))
    # fmt: on
    return metric_rc



"""."""

from kinematic_tracker.nd.derivative_sorted import DerivativeSorted
from kinematic_tracker.nd.gen_x import NdKkfMatGenX


def test_init_and_get_num_x() -> None:
    """."""
    gen_x = NdKkfMatGenX([3, 1], [3, 2])
    der_sorted = DerivativeSorted(gen_x)
    assert der_sorted.get_num_x() == 11



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_upd_for_loose_tracks(tracker: NdKkfTracker) -> None:
    """."""
    tracker.update_for_loose_tracks({}, 0)
    assert tracker.tracks == []



"""
This module defines the `MatchDriverBase` class, which provides a common
functionality for validating the shapes of the input metric and storing result
of the association.
"""

import numpy as np


class MatchDriverBase:
    """
    A base class for managing associations between reports and targets.

    Attributes:
        num_reports_max: The maximum number of reports allowed.
        num_targets_max: The maximum number of targets allowed.
        reports: An integer array storing the indices of matched reports (result).
        targets: An integer array storing the indices of matched targets (result).
        num_matches: The current number of matches.
    """

    def __init__(self, num_rows_max: int, num_cols_max: int) -> None:
        """
        Initializes the MatchDriverBase instance.

        Args:
            num_rows_max: The maximum number of rows (reports).
            num_cols_max: The maximum number of columns (targets).
        """
        self.num_reports_max = num_rows_max
        self.num_targets_max = num_cols_max
        num_matches_max = min(num_rows_max, num_cols_max)
        self.reports = np.zeros(num_matches_max, dtype=int)
        self.targets = np.zeros(num_matches_max, dtype=int)
        self.num_matches = 0

    def check_max_shape(self, shape: tuple[int, int]) -> None:
        """
        Validates that the given shape does not exceed the maximum allowed dimensions.

        Args:
            shape (tuple[int, int]): A tuple representing the shape (rows, columns).

        Raises:
            ValueError: If the number of rows exceeds `num_reports_max` or
                        the number of columns exceeds `num_targets_max`.
        """
        if shape[0] > self.num_reports_max:
            raise ValueError(f'Too much reports? {shape[0]} {self.num_reports_max}')
        if shape[1] > self.num_targets_max:
            raise ValueError(f'Too much targets? {shape[1]} {self.num_targets_max}')

    def get_reports_targets(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Retrieves the matched reports and targets.

        Returns:
            tuple[np.ndarray, np.ndarray]: A tuple containing two arrays:
                                           - The (valid chunk of) matched reports.
                                           - The (valid chunk of) matched targets.
        """
        return self.reports[: self.num_matches], self.targets[: self.num_matches]



"""."""

import numpy as np

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_correct_associated(tracker: NdKkfTracker) -> None:
    """."""
    r2z = [np.ones((6, 1))]
    r2id = [567]
    score_rt = np.zeros((1, 0))
    tracker.correct_associated([], [], score_rt, r2z, r2id)
    assert tracker.tracks == []
